import ctypes
import platform

from conan_explorer.settings import GUI_STYLE_FLUENT
from conan_explorer.ui.common import ThemedWidget

from . import (
    LEFT_MENU_MAX_WIDTH,
    LEFT_MENU_MIN_WIDTH,
    RIGHT_MENU_MAX_WIDTH,
    RIGHT_MENU_MIN_WIDTH,
    gen_obj_name,
)

if platform.system() == "Windows":
    from ctypes.wintypes import MSG

from enum import Enum
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Type, TypeVar, Union

from PySide6.QtCore import (
    QByteArray,
    QEasingCurve,
    QEvent,
    QObject,
    QPoint,
    QPropertyAnimation,
    QRect,
    QSize,
    Qt,
)
from PySide6.QtGui import QAction, QHoverEvent, QMouseEvent
from PySide6.QtWidgets import QMainWindow, QPushButton, QSizePolicy, QWidget
from typing_extensions import override

# uses Logger, settings and theming related functions
from conan_explorer import AUTOCLOSE_SIDE_MENU
from conan_explorer.app.system import is_windows_11

from ..common import get_themed_asset_icon
from .side_menu import SideSubMenu

if TYPE_CHECKING:
    from ..plugin.handler import PluginInterfaceV1


class WidgetNotFoundException(Exception):
    """Raised, when a widget searched for, ist in the parent container."""

    pass


class ResizeDirection(Enum):
    default = 0
    top = 1
    left = 2
    right = 3
    bottom = 4
    top_left = 5
    top_right = 6
    bottom_left = 7
    bottom_right = 8


class FluentWindow(QMainWindow, ThemedWidget):
    MENU_ANIM_DURATION = 0  # disable for performance

    class PageStore:
        """Saves all relevant information for pages accessible from the left menu
        and provides easy retrieval methods for all members."""

        def __init__(self) -> None:
            self._page_widgets: Dict[
                str, Tuple[QPushButton, QWidget, Optional[SideSubMenu], str]
            ] = {}

        def get_button_by_name(self, name: str) -> QPushButton:
            return self._page_widgets[gen_obj_name(name)][0]

        def get_page_by_name(self, name: str) -> QWidget:
            return self._page_widgets[gen_obj_name(name)][1]

        def get_side_menu_by_name(self, name: str) -> "Optional[SideSubMenu]":
            return self._page_widgets[gen_obj_name(name)][2]

        def get_display_name_by_name(self, name: str) -> str:
            return self._page_widgets[gen_obj_name(name)][3]

        def get_side_menu_by_type(self, type_name: Type) -> "Optional[SideSubMenu]":
            for _, (_, page, menu, _) in self._page_widgets.items():
                if page.__class__.__name__ == type_name.__name__:
                    return menu
            raise WidgetNotFoundException(f"{type_name} not in page_widgets!")

        def get_button_by_type(self, type_name: Type) -> QPushButton:
            for _, (button, page, _, _) in self._page_widgets.items():
                if page.__class__.__name__ == type_name.__name__:
                    return button
            raise WidgetNotFoundException(f"{type_name} not in page_widgets!")

        T = TypeVar("T")

        def get_page_by_type(self, type_name: Type[T]) -> T:
            for (
                _,
                (_, page, _, _),
            ) in self._page_widgets.items():
                if page.__class__.__name__ == type_name.__name__:
                    return page  # type: ignore
            raise WidgetNotFoundException(f"{type_name} not in page_widgets!")

        def get_all_buttons(self):
            buttons = []
            for button, _, _, _ in self._page_widgets.values():
                buttons.append(button)
            return buttons

        def get_all_pages(self) -> List["PluginInterfaceV1"]:
            pages = []
            for (
                _,
                (_, page, _, _),
            ) in self._page_widgets.items():
                pages.append(page)
            return pages

        def add_new_page(self, name: str, button, page, right_sub_menu):
            self._page_widgets[gen_obj_name(name)] = (button, page, right_sub_menu, name)

        def remove_page_by_name(self, name: str):
            button, widget, menu, _ = self._page_widgets.pop(gen_obj_name(name))
            button.hide()
            widget.hide()
            button.deleteLater()
            if menu:
                menu.deleteLater()
            widget.deleteLater()

        def remove_page_extras_by_name(self, name: str):
            """Remove page but not the widget itself, in case it is managed
            by another mechanism like plugins"""
            try:
                button, _, menu, _ = self._page_widgets.pop(gen_obj_name(name))
                button.hide()
                button.deleteLater()
                if menu:
                    menu.deleteLater()
            except Exception:  # fail silently - nothing to remove
                pass

    def __init__(self, title_text: str = "", native_windows_fcns=True):
        QMainWindow.__init__(self)
        ThemedWidget.__init__(self, None)
        from .fluent_window_ui import Ui_MainWindow  # need to resolve circular import

        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

        # Window style setup
        wt = Qt.WindowType
        self.setWindowFlags(
            wt.FramelessWindowHint
            | wt.WindowSystemMenuHint  # type: ignore
            | wt.WindowMinimizeButtonHint
            | wt.WindowMaximizeButtonHint
        )
        if (
            is_windows_11() or platform.system() == "Linux"
        ):  # To hide black edges around the border rounding
            self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground, True)

        self._use_native_windows_fcns = (
            True if platform.system() == "Windows" and native_windows_fcns else False
        )

        # all buttons and widgets to be able to shown on the main page
        # (from settings and left menu)
        self.page_widgets = FluentWindow.PageStore()

        # Right bottom general settings Sub Menu
        self.main_general_settings_menu = SideSubMenu(
            self.ui.right_menu_bottom_content_sw, "General Settings", True
        )
        self.ui.right_menu_bottom_content_sw.addWidget(self.main_general_settings_menu)
        self.ui.right_menu_bottom_content_sw.setCurrentWidget(self.main_general_settings_menu)

        # resize related variables
        self._resize_direction = ResizeDirection.default
        self._resize_point = QPoint()
        self._last_geometry = QRect()
        self.title_text = title_text
        self._drag_position: Optional[QPoint] = None

        self.ui.left_menu_frame.setMinimumWidth(LEFT_MENU_MIN_WIDTH)
        self.ui.left_menu_frame.setMaximumWidth(LEFT_MENU_MIN_WIDTH)
        menu_margins = self.ui.left_menu_bottom_subframe.layout().contentsMargins()
        button_offset = menu_margins.right() + menu_margins.left()

        # fix buttons sizes, so they don't expand on toggling the menu
        self.ui.toggle_left_menu_button.setMinimumWidth(LEFT_MENU_MIN_WIDTH - button_offset)
        self.ui.toggle_left_menu_button.setMaximumWidth(LEFT_MENU_MIN_WIDTH - button_offset)

        self.ui.settings_button.setFixedWidth(LEFT_MENU_MIN_WIDTH - button_offset)

        # set all built-in styled icons
        self.set_themed_icon(self.ui.toggle_left_menu_button, "icons/menu_stripes.svg")
        self.set_themed_icon(self.ui.settings_button, "icons/settings.svg")
        self.set_themed_icon(self.ui.minimize_button, "icons/minus.svg")
        self.set_themed_icon(self.ui.close_button, "icons/close.svg")
        qss_icons = [
            "icons/check_box_empty.svg",
            "icons/check_box_checked.svg",
            "icons/forward.svg",
            "icons/expand.svg",
        ]
        for icon in qss_icons:
            get_themed_asset_icon(icon, force_dark_mode=True)

        from ..common.theming import get_gui_style

        style = get_gui_style()
        if style == GUI_STYLE_FLUENT:
            self.ui.close_button.setIconSize(QSize(16, 16))

        # window buttons
        self.ui.restore_max_button.clicked.connect(self.on_maximize_restore)
        self.ui.minimize_button.clicked.connect(self.on_minimize)
        self.ui.close_button.clicked.connect(self.close)

        self.ui.left_menu_top_subframe.mouseMoveEvent = self.move_window
        self.ui.top_frame.mouseMoveEvent = self.move_window
        self.ui.top_frame.mouseDoubleClickEvent = self.on_maximize_restore

        self.ui.toggle_left_menu_button.clicked.connect(self.toggle_left_menu)
        self.ui.settings_button.clicked.connect(self.toggle_right_menu)

        # clear default strings
        self.ui.title_label.setText("")
        self.ui.title_icon_label.hide()

        # initial maximize state
        self.set_restore_max_button_state()
        self.enable_windows_native_animations()
        self.installEventFilter(self)  # used for resizing

        # setup context menu for console
        self.ui.console.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.ui.console.customContextMenuRequested.connect(
            self.on_console_context_menu_requested
        )
        self._console_context_menu = self.ui.console.createStandardContextMenu()
        clear_console_action = QAction("Clear log", self)
        self.set_themed_icon(clear_console_action, "icons/clear_all.svg")
        self._console_context_menu.addAction(clear_console_action)
        clear_console_action.triggered.connect(self.ui.console.clear)

    def on_console_context_menu_requested(self, position):
        self._console_context_menu.exec(self.ui.console.mapToGlobal(position))

    @override
    def nativeEvent(self, eventType: Union[QByteArray, bytes], message: int) -> object:
        """Platform native events"""
        retval = QMainWindow.nativeEvent(self, eventType, message)
        if str(eventType) == "b'windows_generic_MSG'":
            msg = MSG.from_address(message.__int__())  # type: ignore
            # ignore WM_NCCALCSIZE event. Suppresses native Window drawing of title-bar.
            if msg.message == 131:
                return True, 0
        return retval

    @override
    def mousePressEvent(self, event: QMouseEvent):
        """Helper for moving window to know mouse position (Non Windows)"""
        self._drag_position = event.globalPosition().toPoint()

    def apply_theme(self):
        "This function must be able to reload all icons from the left and right menu bar."
        self.reload_themed_icons()
        self.set_restore_max_button_state(force=True)
        for submenu in self.ui.right_menu_bottom_content_sw.findChildren(SideSubMenu):
            submenu.reload_themed_icons()  # type: ignore
        for submenu in self.ui.right_menu_top_content_sw.findChildren(SideSubMenu):
            submenu.reload_themed_icons()  # type: ignore

    def move_window(self, event: QMouseEvent):
        # do nothing if the resize function is active
        if self.cursor().shape() != Qt.CursorShape.ArrowCursor:
            return
        if self._use_native_windows_fcns:
            # enables Windows snap functions
            ctypes.windll.user32.ReleaseCapture()
            # emit native move signal with the following constants:
            # WM_SYSCOMMAND = 274, SC_MOVE = 61456 + HTCAPTION = 2
            ctypes.windll.user32.SendMessageA(int(self.window().winId()), 274, 61458, 0)
        else:
            # if maximized, return to normal be able to move
            if self.isMaximized():
                self.on_maximize_restore(None)
            # qt move
            if event.buttons() == Qt.MouseButton.LeftButton:
                if self._drag_position is None:
                    return
                self.move(self.pos() + event.globalPosition().toPoint() - self._drag_position)  # type: ignore
                self._drag_position = event.globalPosition().toPoint()
                event.accept()

    def enable_windows_native_animations(self):
        if platform.system() != "Windows":
            return
        # sets up thickframe and other needed flag for WIN-Arrow functions an animations to
        # work needs custom resizing and border suppresion functions to work
        GWL_STYLE = -16
        WS_MAXIMIZEBOX = 65536
        WS_CAPTION = 12582912
        CS_DBLCLKS = 8
        WS_THICKFRAME = 262144
        WS_BORDER = 8388608
        style = ctypes.windll.user32.GetWindowLongA(int(self.winId()), GWL_STYLE)
        ctypes.windll.user32.SetWindowLongA(
            int(self.winId()),
            GWL_STYLE,
            style | WS_BORDER | WS_THICKFRAME | WS_MAXIMIZEBOX | CS_DBLCLKS | WS_CAPTION,
        )

    def add_left_menu_entry(
        self,
        name: str,
        asset_icon: str,
        is_upper_menu: bool,
        page_widget: QWidget,
        create_page_menu=False,
    ):
        button = QPushButton("", self.ui.left_menu_frame)
        self.set_themed_icon(button, asset_icon)
        button.setObjectName(gen_obj_name(name))
        size_policy = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        size_policy.setHeightForWidth(button.sizePolicy().hasHeightForWidth())
        button.setSizePolicy(size_policy)
        button.setMinimumSize(QSize(64, 50))
        button.setMaximumHeight(50)
        button.setToolTip(name)
        button.setIconSize(QSize(32, 32))
        button.setStyleSheet("text-align:middle;")
        button.setCheckable(True)
        # enable resizing of console on every page
        page_widget.setMinimumHeight(300)

        button.clicked.connect(self.switch_page)

        # add a stacked widget to right menu with title
        right_sub_menu = None
        if create_page_menu:
            right_sub_menu = SideSubMenu(self.ui.right_menu_top_content_sw, name, True)
            self.ui.right_menu_top_content_sw.addWidget(right_sub_menu)
        self.page_widgets.add_new_page(name, button, page_widget, right_sub_menu)

        if is_upper_menu:
            self.ui.left_menu_middle_subframe.layout().addWidget(button)
        else:
            self.ui.left_menu_bottom_subframe.layout().addWidget(button)
        self.ui.page_stacked_widget.addWidget(page_widget)

    # can only be called for top level menu
    def add_right_bottom_menu_main_page_entry(
        self, name: str, page_widget: QWidget, asset_icon: str = ""
    ):
        """Add a main page entry, where the button is in the settings menu"""
        button = self.main_general_settings_menu.add_button_menu_entry(
            name, self.switch_page, asset_icon
        )
        self.page_widgets.add_new_page(name, button, page_widget, None)
        self.ui.page_stacked_widget.addWidget(page_widget)
        button.clicked.connect(self.toggle_right_menu)

    def switch_page(self):
        sender_button = self.sender()
        assert isinstance(
            sender_button, QPushButton
        ), "Switch page can only be triggered from a button!"
        obj_name = sender_button.objectName()
        page = self.page_widgets.get_page_by_name(obj_name)
        # switch page_stacked_widget to the saved page
        self.ui.page_stacked_widget.setCurrentWidget(page)
        # change button stylings - reset old selections and highlight new one
        for button in self.page_widgets.get_all_buttons():
            button.setChecked(False)

        sender_button.setChecked(True)
        self.ui.page_title.setText(self.page_widgets.get_display_name_by_name(obj_name))

        # update page settings view
        side_menu = self.page_widgets.get_side_menu_by_name(obj_name)
        if not side_menu:
            self.ui.right_menu_top_content_sw.hide()
        else:
            self.ui.right_menu_top_content_sw.show()
            self.ui.right_menu_top_content_sw.setCurrentWidget(side_menu)
        if AUTOCLOSE_SIDE_MENU and self.ui.settings_button.isChecked():
            self.toggle_right_menu()

    def toggle_left_menu(self):
        width = self.ui.left_menu_frame.width()

        # switch extended and minimized state
        if width < LEFT_MENU_MAX_WIDTH:
            width_to_set = LEFT_MENU_MAX_WIDTH
            maximize = True
        else:
            width_to_set = LEFT_MENU_MIN_WIDTH
            maximize = False

        self.left_anim = QPropertyAnimation(self.ui.left_menu_frame, b"minimumWidth")  # type: ignore
        self.left_anim.setDuration(self.MENU_ANIM_DURATION)
        self.left_anim.setStartValue(width)
        self.left_anim.setEndValue(width_to_set)
        self.left_anim.setEasingCurve(QEasingCurve.Type.InOutQuart)
        self.left_anim.start()

        # hide title
        if maximize:
            self.ui.title_label.setText(self.title_text)
            self.ui.title_icon_label.show()
        else:
            self.ui.title_label.setText("")
            self.ui.title_icon_label.hide()

        # hide menu button texts
        # name, (button, _) in self.page_entries.items():
        for button in self.ui.left_menu_middle_subframe.findChildren(QPushButton):
            if not isinstance(button, QPushButton):
                return
            name = self.page_widgets.get_display_name_by_name(button.objectName())
            if maximize:
                button.setText(name)
                button.setStyleSheet("text-align:left;")
            else:
                button.setText("")
                button.setStyleSheet("text-align:middle;")

        if self.ui.settings_button.isChecked():
            self.toggle_right_menu()

    def toggle_right_menu(self):
        width = self.ui.right_menu_frame.width()
        if width < RIGHT_MENU_MAX_WIDTH:
            width_to_set = RIGHT_MENU_MAX_WIDTH
        else:
            width_to_set = RIGHT_MENU_MIN_WIDTH
            self.ui.settings_button.setChecked(False)
        self.right_anim = QPropertyAnimation(self.ui.right_menu_frame, b"minimumWidth")  # type: ignore
        self.right_anim.setDuration(self.MENU_ANIM_DURATION)
        self.right_anim.setStartValue(width)
        self.right_anim.setEndValue(width_to_set)
        self.right_anim.setEasingCurve(QEasingCurve.Type.Linear)
        self.right_anim.start()

    @override
    def eventFilter(self, watched: QObject, event: QEvent) -> bool:
        """Implements window resizing"""
        self.set_restore_max_button_state()
        if self.isMaximized():  # no resize when maximized
            # forced reset of cursor
            if self.cursor().shape() != Qt.CursorShape.ArrowCursor:
                self.setCursor(Qt.CursorShape.ArrowCursor)
            return super().eventFilter(watched, event)
        # Use isinstance instead of type because of typehinting
        if isinstance(event, QHoverEvent):
            if event.type() == event.Type.HoverMove:
                # cursor position control for cursor shape setup
                self.handle_resize_cursor(event)
        elif isinstance(event, QMouseEvent):
            if event.type() == event.Type.MouseButtonPress:
                if event.button() != Qt.MouseButton.LeftButton:
                    return super().eventFilter(watched, event)
                # resizing
                if self.cursor().shape() != Qt.CursorShape.ArrowCursor:
                    # save the starting point of resize
                    self._resize_point = self.mapToGlobal(event.pos())
                    self._last_geometry = self.geometry()
                    self.resizing()
            else:  # Hacky fix for when the cursor is not reset after resizing
                self.setCursor(Qt.CursorShape.ArrowCursor)

        return super().eventFilter(watched, event)

    def handle_resize_cursor(self, event: QHoverEvent, x_offset=10, y_offset=8):
        # using relative position, since the event can only be fired inside of the Window
        rect = self.rect()  # rect()
        top_left = rect.topLeft()
        top_right = rect.topRight()
        bottom_left = rect.bottomLeft()
        bottom_right = rect.bottomRight()
        position = event.position().toPoint()  # relative pos to window
        width = self.width()
        height = self.height()
        cs = Qt.CursorShape
        if QRect(
            top_left.x() + x_offset, top_left.y(), width - 2 * x_offset, y_offset
        ).contains(position):
            self._resize_direction = ResizeDirection.top
            self.setCursor(cs.SizeVerCursor)
        elif QRect(
            bottom_left.x() + x_offset, bottom_left.y(), width - 2 * x_offset, -y_offset
        ).contains(position):
            self._resize_direction = ResizeDirection.bottom
            self.setCursor(cs.SizeVerCursor)
        elif QRect(
            top_right.x() - x_offset,
            top_right.y() + y_offset,
            x_offset,
            height - 2 * y_offset,
        ).contains(position):
            self._resize_direction = ResizeDirection.right
            self.setCursor(cs.SizeHorCursor)
        elif QRect(
            top_left.x() + x_offset,
            top_left.y() + y_offset,
            -x_offset,
            height - 2 * y_offset,
        ).contains(position):
            self._resize_direction = ResizeDirection.left
            self.setCursor(cs.SizeHorCursor)
        elif QRect(top_right.x(), top_right.y(), -x_offset, y_offset).contains(position):
            self._resize_direction = ResizeDirection.top_right
            self.setCursor(cs.SizeBDiagCursor)
        elif QRect(bottom_left.x(), bottom_left.y(), x_offset, -y_offset).contains(position):
            self._resize_direction = ResizeDirection.bottom_left
            self.setCursor(cs.SizeBDiagCursor)
        elif QRect(top_left.x(), top_left.y(), x_offset, y_offset).contains(position):
            self._resize_direction = ResizeDirection.top_left
            self.setCursor(cs.SizeFDiagCursor)
        elif QRect(bottom_right.x(), bottom_right.y(), -x_offset, -y_offset).contains(position):
            self._resize_direction = ResizeDirection.bottom_right
            self.setCursor(cs.SizeFDiagCursor)
        else:  # no resize
            self._resize_direction = ResizeDirection.default
            self.setCursor(cs.ArrowCursor)

    def resizing(self):
        ed = Qt.Edge
        window = self.window().windowHandle()
        if self._resize_direction == ResizeDirection.top:
            window.startSystemResize(ed.TopEdge)
        elif self._resize_direction == ResizeDirection.bottom:
            window.startSystemResize(ed.BottomEdge)
        elif self._resize_direction == ResizeDirection.right:
            window.startSystemResize(ed.RightEdge)
        elif self._resize_direction == ResizeDirection.left:
            window.startSystemResize(ed.LeftEdge)
        elif self._resize_direction == ResizeDirection.top_right:
            window.startSystemResize(ed.TopEdge | ed.RightEdge)
        elif self._resize_direction == ResizeDirection.bottom_right:
            window.startSystemResize(ed.BottomEdge | ed.RightEdge)
        elif self._resize_direction == ResizeDirection.bottom_left:
            window.startSystemResize(ed.BottomEdge | ed.LeftEdge)
        elif self._resize_direction == ResizeDirection.top_left:
            window.startSystemResize(ed.TopEdge | ed.LeftEdge)

    def on_maximize_restore(self, event=None):  # dummy arg to be used as an event slot
        if self.isMaximized():
            self.setWindowState(Qt.WindowState.WindowNoState)
            self.showNormal()
        else:
            self.setWindowState(Qt.WindowState.WindowMaximized)
            self.showMaximized()

    def on_minimize(self, event=None):
        self.setWindowState(Qt.WindowState.WindowMinimized)
        self.showMinimized()

    def set_restore_max_button_state(self, force=False):
        """This toggles the restore/maximize button in the top right button cluster"""
        if self.isMaximized():
            if self.ui.restore_max_button.icon().themeName() == "restore" and not force:
                return
            icon = get_themed_asset_icon("icons/restore.svg")
            icon.setThemeName("restore")
            self.ui.restore_max_button.setIcon(icon)
        else:
            if self.ui.restore_max_button.icon().themeName() == "maximize" and not force:
                return
            icon = get_themed_asset_icon("icons/maximize.svg")
            icon.setThemeName("maximize")
            self.ui.restore_max_button.setIcon(icon)



from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union
from typing_extensions import Self
from abc import abstractmethod
from conan_explorer import INVALID_CONAN_REF, INVALID_PATH, conan_version
from conan_explorer.app.logger import Logger
from .types import (ConanAvailableOptions,  ConanPkg, ConanRef, ConanPkgRef, 
            ConanOptions, ConanPackageId, ConanPackagePath, ConanSettings, EditablePkg, Remote)

if TYPE_CHECKING:
    from conan_explorer.conan_wrapper.conan_cache import ConanInfoCache

#### Interface
class ConanUnifiedApi():
    """ 
    API abstraction to provide compatibility between ConanV1 and V2 APIs. 
    """

    @abstractmethod
    def __init__(self):
        self.info_cache: "ConanInfoCache"
        ...

    @abstractmethod
    def init_api(self) -> Self:
        """ Instantiate the internal Conan api. """
        raise NotImplementedError

### General commands ###

    @abstractmethod
    def remove_locks(self):
        """ Remove local cache locks (Currently for V1 only) """
        raise NotImplementedError

    @abstractmethod
    def get_profiles(self) -> List[str]:
        """ Return a list of all profiles """
        raise NotImplementedError

    @abstractmethod
    def get_profile_settings(self, profile_name: str) -> ConanSettings:
        """ Return a dict of settings for a profile """
        raise NotImplementedError

    @abstractmethod
    def get_default_settings(self) -> ConanSettings:
        """
        Return the settings of the conan default profile. 
        This is not necessarily the same as get_profile_settings("default"), 
        because its content can be changed.
        """
        raise NotImplementedError

    @abstractmethod
    def get_profiles_path(self) -> Path:
        """ Get the path to the folder where profiles are located"""
        raise NotImplementedError

    @abstractmethod
    def get_remote_user_info(self, remote_name: str) -> Tuple[str, bool]: # username, is_authenticated
        """ Get username and authenticated info for a remote. """
        raise NotImplementedError

    @abstractmethod
    def get_settings_file_path(self) -> Path:
        """ Return conan settings file path (settings.yml) """
        raise NotImplementedError

    @abstractmethod
    def get_config_file_path(self) -> Path:
        """ Return conan config file path (conan.conf) """
        raise NotImplementedError

    @abstractmethod
    def get_config_entry(self, config_name: str, default_value: Any) -> Any:
        """ Return a conan config entry value (conan.conf). 
        Use default_value for non existing values. """
        raise NotImplementedError

    @abstractmethod
    def get_revisions_enabled(self) -> bool:
        """ Return if revisions are enabled for Conan V1. Always true in V2 mode. """
        raise NotImplementedError

    @abstractmethod
    def get_user_home_path(self) -> Path:
        """ Return Conan user home path, where e.g. settings reside """
        raise NotImplementedError

    @abstractmethod
    def get_storage_path(self) -> Path:
        """ Return Conan storage path, where packages are saved """
        raise NotImplementedError

    @abstractmethod
    def get_short_path_root(self) -> Path:
        """ Return short path root for Windows for Conan V1. 
        Sadly there is no built-in way to do so. """
        raise NotImplementedError

    @abstractmethod
    def get_package_folder(self, conan_ref: ConanRef, package_id: str) -> ConanPackagePath:
        " Get the fully resolved pkg path from the ref and the specific package (id) "
        raise NotImplementedError

    @abstractmethod
    def get_export_folder(self, conan_ref: ConanRef) -> Path:
        """ Get the export folder from a reference. Will throw for editables. """
        raise NotImplementedError

    @abstractmethod
    def get_conanfile_path(self, conan_ref: ConanRef) -> Path:
        """ Get local conanfile path. If it is not localy available, download it."""
        raise NotImplementedError

### Remotes 

    @abstractmethod
    def get_remotes(self, include_disabled=False) -> List[Remote]:
        """ Return a list of all remotes. """
        raise NotImplementedError

    @abstractmethod
    def add_remote(self, remote_name: str, url: str, verify_ssl: bool):
        """  Add a new remote with the selected options. Disabled is always false. """
        raise NotImplementedError
    
    @abstractmethod
    def rename_remote(self, remote_name: str, new_name: str):
        """ Rename a remote """
        raise NotImplementedError

    @abstractmethod
    def remove_remote(self, remote_name: str):
        """  Remove a remote """
        raise NotImplementedError
    
    @abstractmethod
    def update_remote(self, remote_name: str, url: str, verify_ssl: bool, 
                      disabled: bool, index: Optional[int]):
        """ Update a remote with new information and reorder with index  """
        raise NotImplementedError
    
    @abstractmethod
    def disable_remote(self, remote_name: str, disabled: bool):
        """ Enable or disable a remote """
        raise NotImplementedError

    @abstractmethod
    def login_remote(self, remote_name: str, user_name: str, password: str):
        """ Login to a remote with credentials 
        This method will throw if wrong credentials are entered, 
        so we can catch the first error, when logging into multiple remotes
        and do not retry, possibly locking the user.
        """
        raise NotImplementedError

### Install related methods ###


    @abstractmethod
    def install_reference(self, conan_ref: ConanRef, 
                          conan_settings: Optional[ConanSettings]=None,
                          conan_options: Optional[ConanOptions]=None, profile="", 
                          update=True, quiet=False, generators: List[str] = []
                          ) -> Tuple[ConanPackageId, ConanPackagePath]:
        """
        Try to install a conan reference (without id) with the provided extra information.
        Uses plain conan install (No auto determination of best matching package)
        Returns the actual pkg_id and the package path.
        """
        raise NotImplementedError

    @abstractmethod
    def install_package(self, conan_ref: ConanRef, package: ConanPkg,
                            update=True) -> Tuple[ConanPackageId, ConanPackagePath]:
        """
        Try to install a conan package (id) with the provided extra information.
        Returns the installed id and a valid pkg path, if installation was succesfull.
        WARNING: The installed id can differ from the requested one, 
        because there is no built-in way in conan to install a specific package id!
        """
        raise NotImplementedError

    @abstractmethod
    def get_path_or_auto_install(self, conan_ref: ConanRef, 
            conan_options: Optional[ConanOptions]=None, 
            update=False) -> Tuple[ConanPackageId, ConanPackagePath]:
        """ Return the pkg_id and package folder of a conan reference 
        and auto-install it with the best matching package, if it is not available """
        raise NotImplementedError

    @abstractmethod
    def install_best_matching_package(self, conan_ref: ConanRef, 
            conan_options: Optional[ConanOptions]=None,
            update=False) -> Tuple[ConanPackageId, ConanPackagePath]:
        raise NotImplementedError

    @abstractmethod
    def get_options_with_default_values(self, 
            conan_ref: ConanRef) -> Tuple[ConanAvailableOptions, ConanOptions]:
        """ Return the available options and their default values as dict."""
        raise NotImplementedError


### Local References and Packages ###

    @abstractmethod
    def get_conan_buildinfo(self, conan_ref: ConanRef, conan_settings: ConanSettings,
                            conan_options: Optional[ConanOptions]=None) -> str:
        """ Read conan buildinfo and return as string """
        raise NotImplementedError

    @abstractmethod
    def get_editable(self, conan_ref: Union[ConanRef, str]) -> EditablePkg:
        """ Get an editable object from conan reference. """
        raise NotImplementedError

    @abstractmethod
    def get_editables_package_path(self, conan_ref: ConanRef) -> Path:
        """ Get package path of an editable reference. """
        raise NotImplementedError
    
    @abstractmethod
    def get_editables_output_folder(self, conan_ref: ConanRef) -> Optional[Path]:
        """ Get output folder of an editable reference. """
        raise NotImplementedError

    @abstractmethod
    def get_editable_references(self) -> List[ConanRef]:
        """ Get all local editable references. """
        raise NotImplementedError

    @abstractmethod
    def add_editable(self, conan_ref: Union[ConanRef, str], path: str, output_folder: str) -> bool:
        """ Add an editable reference. """
        raise NotImplementedError

    @abstractmethod
    def remove_editable(self, conan_ref: Union[ConanRef, str]) -> bool:
        """ Remove an editable reference. """
        raise NotImplementedError

    @abstractmethod
    def remove_reference(self, conan_ref: ConanRef, pkg_id: str=""):
        """ Remove a conan reference and it's package if specified via id """
        raise NotImplementedError

    @abstractmethod
    def find_best_matching_local_package(self, conan_ref: ConanRef, 
            conan_options: Optional[ConanOptions]=None) -> ConanPkg:
        """ Find a package in the local cache """
        raise NotImplementedError

    @abstractmethod
    def get_best_matching_local_package_path(self, conan_ref: ConanRef, 
            conan_options: Optional[ConanOptions]=None
            ) -> Tuple[ConanPackageId, ConanPackagePath]:
        " Return the pkg_id and pkg folder of a conan reference, if it is installed. "
        raise NotImplementedError

    @abstractmethod
    def get_all_local_refs(self) -> List[ConanRef]:
        """ Returns all locally installed conan references """
        raise NotImplementedError

    @abstractmethod
    def get_local_pkgs_from_ref(self, conan_ref: ConanRef) -> List[ConanPkg]:
        """ Returns all installed pkg ids for a reference. """
        raise NotImplementedError

    def get_local_pkg_from_id(self, pkg_ref: ConanPkgRef) -> ConanPkg:
        """ Returns an installed pkg from reference and id """
        raise NotImplementedError

    def get_local_pkg_from_path(self, conan_ref: ConanRef, path: Path):
        """ For reverse lookup - give info from path """
        raise NotImplementedError

### Remote References and Packages ###

    @abstractmethod
    def search_recipes_in_remotes(self, query: str, remote_name="all") -> List[ConanRef]:
        """ Search in all remotes for a specific query. 
        Returns a list if unqiue and ordered ConanRefs. """
        raise NotImplementedError

    @abstractmethod
    def search_recipe_all_versions_in_remotes(self, conan_ref: ConanRef) -> List[ConanRef]:
        """ Search in all remotes for all versions of a conan ref """
        raise NotImplementedError

    @abstractmethod
    def get_remote_pkgs_from_ref(self, conan_ref: ConanRef, remote: Optional[str],
                                 query=None) -> List[ConanPkg]:
        " Return all packages for a reference in a specific remote with an optional query. "
        raise NotImplementedError

    @abstractmethod
    def get_remote_pkg_from_id(self, pkg_ref: ConanPkgRef) -> ConanPkg:
        """ Returns a remote pkg from reference and id """
        raise NotImplementedError

    @abstractmethod
    def find_best_matching_package_in_remotes(self, conan_ref: ConanRef, 
            conan_options: Optional[ConanOptions]) -> List[ConanPkg]:
        """ Find a package with options in the remotes """
        raise NotImplementedError

    @abstractmethod
    def find_best_matching_packages(self, conan_ref: ConanRef,
                                conan_options: Optional[ConanOptions] = None, 
                                remote_name: Optional[str] = None) -> List[ConanPkg]:
        """
        This method tries to find the best matching packages either locally or in a remote,
        based on the users machine and the supplied options.
        """
        raise NotImplementedError

    @staticmethod
    def generate_canonical_ref(conan_ref: ConanRef) -> str:
        "Creates a full ref from a short ref, e.g. product/1.0.0 -> product/1.0.0@_/_"
        raise NotImplementedError

    @staticmethod
    def build_conan_profile_name_alias(conan_settings: ConanSettings) -> str:
        "Build a human readable pseduo profile name, like Windows_x64_vs16_v142_release"
        raise NotImplementedError


##### Common functions - these use either only the version specific base commands 
# or do not depend on any version specifics at all
class ConanCommonUnifiedApi(ConanUnifiedApi):
    """ 
    High level functions, which use only other ConanUnifiedApi functions are 
    implemented here.
    """

    def __init__(self):
        # no direct Conan API access!
        self.info_cache: "ConanInfoCache"

### Install related methods ###

    def install_package(self, conan_ref: ConanRef, package: ConanPkg, 
                        update=True) -> Tuple[ConanPackageId, ConanPackagePath]:
        """
        Try to install a conan package (id) with the provided extra information.
        Returns the installed id and a valid package path, if installation was succesfull.
        WARNING: The installed id can differ from the requested one, because there is no built-in 
        way in conan to install a specific package id!
        """
        from conans.errors import ConanException
        package_id = package.get("id", "")
        options = package.get("options", {})
        settings = package.get("settings", {})
        Logger().info(
            f"Installing '<b>{str(conan_ref)}</b>':{package_id} with settings: {str(settings)}, "
            f"options: {str(options)} and update={update}\n")
        try:
            installed_id, package_path = self.install_reference(
                conan_ref, update=update, conan_settings=settings, conan_options=options, quiet=True)
            if installed_id != package_id:
                Logger().warning(f"Installed {installed_id} instead of selected {package_id}."
                    "This can happen, if there transitive settings changed in comparison to the build time.")
            return installed_id, package_path
        except ConanException as e:
            Logger().error(f"Can't install package '<b>{str(conan_ref)}</b>': {str(e)}")
            return "", Path(INVALID_PATH)

    def get_path_or_auto_install(self, conan_ref: ConanRef, conan_options: Optional[ConanOptions]=None,
                                 update=False) -> Tuple[ConanPackageId, ConanPackagePath]:
        """ Return the pkg_id and package folder of a conan reference 
        and auto-install it with the best matching package, if it is not available """
        if not update:
            pkg_id, path = self.get_best_matching_local_package_path(conan_ref, conan_options)
            if pkg_id:
                return pkg_id, path
            Logger().info(
                f"'<b>{conan_ref}</b>' with options {repr(conan_options)} is not installed. Searching for packages to install...")

        pkg_id, path = self.install_best_matching_package(conan_ref, conan_options, update=update)
        return pkg_id, path

    def install_best_matching_package(self, conan_ref: ConanRef, 
                            conan_options: Optional[ConanOptions] = None, 
                            update=False) -> Tuple[ConanPackageId, ConanPackagePath]:
        packages: List[ConanPkg] = self.find_best_matching_package_in_remotes(conan_ref, conan_options)
        if not packages:
            self.info_cache.invalidate_remote_package(conan_ref)
            return ("", Path(INVALID_PATH))
        pkg_id, package_path = self.install_package(conan_ref, packages[0], update)
        if package_path.exists():
            return pkg_id, package_path
        return "", Path(INVALID_PATH)

### Local References and Packages ###

    def find_best_matching_local_package(self, conan_ref: ConanRef, 
            conan_options: Optional[ConanOptions]=None) -> ConanPkg:
        """ Find a package in the local cache """
        packages = self.find_best_matching_packages(conan_ref, conan_options, remote_name=None)
        # What to if multiple ones exits? - for now simply take the first entry
        if packages:
            if len(packages) > 1:
                settings = packages[0].get("settings", {})
                pkg_id = packages[0].get("id", "")
                Logger().warning(f"Multiple matching packages found for '<b>{str(conan_ref)}</b>'!\n"
                                 f"Choosing this: {pkg_id} ({self.build_conan_profile_name_alias(settings)})")
            # Update cache with this package
            self.info_cache.update_local_package_path(
                conan_ref, self.get_package_folder(conan_ref, packages[0].get("id", "")))
            return packages[0]
        Logger().debug(
            "No matching local packages found for <b>%s</b>", str(conan_ref))
        return {"id": ""}

    def get_best_matching_local_package_path(self, conan_ref: ConanRef,
            conan_options: Optional[ConanOptions]=None) -> Tuple[ConanPackageId, ConanPackagePath]:
        """ Return the pkg_id and package folder of a conan reference, if it is installed. """
        package = self.find_best_matching_local_package(conan_ref, conan_options)
        if package.get("id", ""):
            return package.get("id", ""), self.get_package_folder(conan_ref, package.get("id", ""))
        return "", Path(INVALID_PATH)

    def get_local_pkg_from_id(self, pkg_ref: ConanPkgRef) -> ConanPkg:
        """ Returns an installed pkg from reference and id """
        package = None
        for package in self.get_local_pkgs_from_ref(pkg_ref.ref):
            if package.get("id", "") == pkg_ref.id:
                return package
        return {"id": ""}

    def get_local_pkg_from_path(self, conan_ref: ConanRef, path: Path):
        """ For reverse lookup - give info from path """
        found_package = None
        for package in self.get_local_pkgs_from_ref(conan_ref):
            if self.get_package_folder(conan_ref, package.get("id", "")) == path:
                found_package = package
                break
        return found_package


### Remote References and Packages ###

    def get_remote_pkg_from_id(self, pkg_ref: ConanPkgRef) -> ConanPkg:
        """ Returns a remote pkg from reference and id """
        package = None
        for remote in self.get_remotes():
            packages = self.get_remote_pkgs_from_ref(pkg_ref.ref, remote.name)
            for package in packages:
                if package.get("id", "") == pkg_ref.id:
                    return package
        return {"id": ""}

    def find_best_matching_package_in_remotes(self, conan_ref: ConanRef, 
                        conan_options: Optional[ConanOptions]=None) -> List[ConanPkg]:
        """ Find a package with options in the remotes """
        for remote in self.get_remotes():
            packages = self.find_best_matching_packages(conan_ref, conan_options, remote.name)
            if packages:
                return packages
        Logger().info(
            f"Can't find a package '<b>{str(conan_ref)}</b>' with options {conan_options} in the <b>remotes</b>")
        return []

    def find_best_matching_packages(self, conan_ref: ConanRef, conan_options: Optional[ConanOptions]=None,
                                    remote_name: Optional[str] = None) -> List[ConanPkg]:
        """
        This method tries to find the best matching packages either locally or in a remote,
        based on the users machine and the supplied options.
        """
        if conan_options is None:
            conan_options = {}
        # skip search on default invalid recipe
        if str(conan_ref) == INVALID_CONAN_REF:
            return []
        found_pkgs: List[ConanPkg] = []
        default_settings: ConanSettings = {}
        try:
            # dynamic prop is ok in try-catch
            default_settings = self.get_default_settings()
            query = f"(arch=None OR arch={default_settings.get('arch')})" \
                    f" AND (os=None OR os={default_settings.get('os')})"
            if conan_version.major == 1:
                query += f" AND (arch_build=None OR arch_build={default_settings.get('arch_build')})" \
                         f" AND (os_build=None OR os_build={default_settings.get('os_build')})"
            found_pkgs = self.get_remote_pkgs_from_ref(conan_ref, remote_name, query)
        except Exception:  # no problem, next
            return []

        # remove debug releases
        no_debug_pkgs = list(filter(lambda pkg: pkg.get("settings", {}).get(
            "build_type", "").lower() != "debug", found_pkgs))
        # check, if a package remained and only then take the result
        if no_debug_pkgs:
            found_pkgs = no_debug_pkgs

        # filter the found packages by the user options
        if conan_options:
            found_pkgs = list(filter(lambda pkg: conan_options.items() <= pkg.get("options", {}).items(), found_pkgs))
            if not found_pkgs:
                return found_pkgs
        # get a set of existing options and reduce default options with them
        min_opts_set = set(map(lambda pkg: frozenset(tuple(pkg.get("options", {}).keys())), found_pkgs))
        min_opts_list = frozenset()
        if min_opts_set:
            min_opts_list = min_opts_set.pop()

        # this calls external code of the recipe
        _, default_options = self.get_options_with_default_values(conan_ref)

        if default_options:
            default_options = dict(filter(lambda opt: opt[0] in min_opts_list, default_options.items()))
            # patch user input into default options to combine the two
            default_options.update(conan_options)
            # convert vals to string
            default_str_options: Dict[str, str] = dict([key, str(value)]
                                            for key, value in default_options.items())
            if len(found_pkgs) > 1:
                comb_opts_pkgs = list(filter(lambda pkg: default_str_options.items() <=
                                             pkg.get("options", {}).items(), found_pkgs))
                if comb_opts_pkgs:
                    found_pkgs = comb_opts_pkgs

        # now we have all matching packages, but with potentially different compilers
        # reduce with default settings
        if len(found_pkgs) > 1:
            same_comp_pkgs = list(filter(lambda pkg: 
                              default_settings.get("compiler", "") ==
                              pkg.get("settings", {}).get("compiler", ""), found_pkgs))
            if same_comp_pkgs:
                found_pkgs = same_comp_pkgs

            same_comp_version_pkgs = list(filter(lambda pkg:
                    default_settings.get("compiler.version", "") ==
                    pkg.get("settings", {}).get("compiler.version", ""), found_pkgs))
            if same_comp_version_pkgs:
                found_pkgs = same_comp_version_pkgs
        return found_pkgs

### Helper methods ###

    @staticmethod
    def _resolve_default_options(default_options_raw: Any) -> ConanOptions:
        """ Default options can be a a dict or name=value as string, or a tuple of it """
        default_options: Dict[str, Any] = {}
        if default_options_raw and isinstance(default_options_raw, str):
            default_option_str = default_options_raw.split("=")
            default_options.update({default_option_str[0]: default_option_str[1]})
        elif default_options_raw and isinstance(default_options_raw, (list, tuple)):
            for default_option in default_options_raw:
                default_option_str = default_option.split("=")
                default_options.update({default_option_str[0]: default_option_str[1]})
        else:
            default_options = default_options_raw
        return default_options

    @staticmethod
    def generate_canonical_ref(conan_ref: ConanRef) -> str:
        if conan_ref.user is None and conan_ref.channel is None:
            return str(conan_ref) + "@_/_"
        return str(conan_ref)

    @staticmethod
    def build_conan_profile_name_alias(conan_settings: ConanSettings) -> str:
        if not conan_settings:
            return "No Settings"

        os = conan_settings.get("os", "")
        if not os:
            os = conan_settings.get("os_target", "")
            if not os:
                os = conan_settings.get("os_build", "")

        arch = conan_settings.get("arch", "")
        if not arch:
            arch = conan_settings.get("arch_target", "")
            if not arch:
                arch = conan_settings.get("arch_build", "")
        if arch == "x86_64":  # shorten x64
            arch = "x64"

        comp = conan_settings.get("compiler", "")
        if comp == "Visual Studio":
            comp = "vs"
        comp_ver = conan_settings.get("compiler.version", "")
        comp_text = comp.lower() + comp_ver.lower()

        comp_toolset = conan_settings.get("compiler.toolset", "")

        bt = conan_settings.get("build_type", "")

        alias = os
        for item in [arch.lower(), comp_text, comp_toolset.lower(), bt.lower()]:
            if item:
                alias += "_" + item

        return alias

    def get_remotes_from_same_server(self, remote: Remote) -> List[Remote]:
        """
        Pass in a remote and return all other remotes with the same base url.
        Currently only for artifactory links.
        """
        remote_groups = self._get_remote_groups()
        for remotes in remote_groups.values():
            for check_remote in remotes:
                if check_remote == remote:
                    return remotes
        return [remote]

    def _get_remote_groups(self) -> Dict[str, List[Remote]]:
        """
        Try to group similar URLs(currently only for artifactory links) 
        and return them in a dict grouped by the full URL.
        """
        remote_groups: Dict[str, List[Remote]] = {}
        for remote in self.get_remotes(include_disabled=True):
            if "artifactory" in remote.url:
                # try to determine root address
                possible_base_url = "/".join(remote.url.split("/")[0:3])
                if not remote_groups.get(possible_base_url):
                    remote_groups[possible_base_url] = [remote]
                else:
                    remotes = remote_groups[possible_base_url]
                    remotes.append(remote)
                    remote_groups.update({possible_base_url: remotes})
            else:
                remote_groups[remote.url] = [remote]
        return remote_groups



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'side_menu.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QFrame, QHBoxLayout, QLabel,
    QLayout, QPushButton, QSizePolicy, QSpacerItem,
    QVBoxLayout, QWidget)

class Ui_SideMenu(object):
    def setupUi(self, SideMenu):
        if not SideMenu.objectName():
            SideMenu.setObjectName(u"SideMenu")
        SideMenu.resize(179, 182)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(SideMenu.sizePolicy().hasHeightForWidth())
        SideMenu.setSizePolicy(sizePolicy)
        self.main_layout = QVBoxLayout(SideMenu)
        self.main_layout.setSpacing(2)
        self.main_layout.setObjectName(u"main_layout")
        self.main_layout.setSizeConstraint(QLayout.SetDefaultConstraint)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.side_menu_title_frame = QFrame(SideMenu)
        self.side_menu_title_frame.setObjectName(u"side_menu_title_frame")
        self.side_menu_title_frame.setFrameShape(QFrame.NoFrame)
        self.title_layout = QHBoxLayout(self.side_menu_title_frame)
        self.title_layout.setSpacing(4)
        self.title_layout.setObjectName(u"title_layout")
        self.title_layout.setSizeConstraint(QLayout.SetNoConstraint)
        self.title_layout.setContentsMargins(-1, 4, 0, 8)
        self.side_menu_title_button = QPushButton(self.side_menu_title_frame)
        self.side_menu_title_button.setObjectName(u"side_menu_title_button")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.side_menu_title_button.sizePolicy().hasHeightForWidth())
        self.side_menu_title_button.setSizePolicy(sizePolicy1)
        self.side_menu_title_button.setMinimumSize(QSize(32, 32))
        self.side_menu_title_button.setMaximumSize(QSize(50, 50))
        self.side_menu_title_button.setFlat(True)

        self.title_layout.addWidget(self.side_menu_title_button)

        self.side_menu_title_label = QLabel(self.side_menu_title_frame)
        self.side_menu_title_label.setObjectName(u"side_menu_title_label")

        self.title_layout.addWidget(self.side_menu_title_label)


        self.main_layout.addWidget(self.side_menu_title_frame)

        self.side_menu_content_frame = QFrame(SideMenu)
        self.side_menu_content_frame.setObjectName(u"side_menu_content_frame")
        self.side_menu_content_frame.setFrameShape(QFrame.NoFrame)
        self.side_menu_content_frame.setFrameShadow(QFrame.Raised)
        self.content_frame_layout = QVBoxLayout(self.side_menu_content_frame)
        self.content_frame_layout.setSpacing(4)
        self.content_frame_layout.setObjectName(u"content_frame_layout")
        self.content_frame_layout.setContentsMargins(0, 0, 9, 0)
        self.side_menu_spacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.content_frame_layout.addItem(self.side_menu_spacer)


        self.main_layout.addWidget(self.side_menu_content_frame)


        self.retranslateUi(SideMenu)

        QMetaObject.connectSlotsByName(SideMenu)
    # setupUi

    def retranslateUi(self, SideMenu):
        SideMenu.setWindowTitle(QCoreApplication.translate("SideMenu", u"Form", None))
        self.side_menu_title_button.setText("")
        self.side_menu_title_label.setText(QCoreApplication.translate("SideMenu", u"TextLabel", None))
    # retranslateUi




from pprint import pformat
from typing import Literal

from dictdiffer import diff
from PySide6.QtCore import QRegularExpression, Qt
from PySide6.QtGui import QAction, QColor, QShowEvent, QTextCharFormat
from PySide6.QtWidgets import QDialog, QListWidgetItem, QMenu
from typing_extensions import override

from conan_explorer.app.logger import Logger
from conan_explorer.conan_wrapper.types import ConanPkg
from conan_explorer.ui.common.syntax_highlighting import ConfigHighlighter
from conan_explorer.ui.common.theming import ThemedWidget


class ConfigDiffHighlighter(ConfigHighlighter):
    """Syntax highlighter to highlight the differences of a dict with different
    background colors (modified, added, removed)"""

    DIFF_NEW_COLOR = QColor("green")
    DIFF_REMOVED_COLOR = QColor("red")
    DIFF_MODIFIED_COLOR = QColor("orange")

    def __init__(self, parent, type: Literal["ini", "yaml"]) -> None:
        super().__init__(parent, type)
        self._reset_diff()

    def _reset_diff(self):
        self.modified_diffs = []
        self.added_diffs = []
        self.removed_diffs = []

    @override
    def highlightBlock(self, text: str):
        super().highlightBlock(text)
        key_format = QTextCharFormat()
        # clear all background color
        key_format.clearBackground()
        expression = QRegularExpression(".*")
        match = expression.match(text)
        self.setFormat(match.capturedStart(), match.capturedLength(), key_format)

        for diff_item in self.modified_diffs:
            key_format.setBackground(self.DIFF_MODIFIED_COLOR)
            expression = QRegularExpression(diff_item)
            match = expression.match(text)
            self.setFormat(match.capturedStart(), match.capturedLength(), key_format)
        for diff_item in self.added_diffs:
            key_format.setBackground(self.DIFF_NEW_COLOR)
            expression = QRegularExpression(diff_item)
            match = expression.match(text)
            self.setFormat(match.capturedStart(), match.capturedLength(), key_format)
        for diff_item in self.removed_diffs:
            key_format.setBackground(self.DIFF_REMOVED_COLOR)
            expression = QRegularExpression(diff_item)
            match = expression.match(text)
            self.setFormat(match.capturedStart(), match.capturedLength(), key_format)


class PkgDiffDialog(QDialog, ThemedWidget):
    def __init__(self, parent) -> None:
        QDialog.__init__(self, parent)
        ThemedWidget.__init__(self, None)
        from .diff_ui import Ui_Dialog

        self._ui = Ui_Dialog()
        self._ui.setupUi(self)
        # TODO test out dialog._ui.left_text_browser.AutoFormattingFlag
        self._init_pkg_context_menu()

        self.setWindowTitle("Compare Packages")
        self.setWindowFlag(Qt.WindowType.WindowMaximizeButtonHint)

        self._left_content = {}
        self._right_content = {}
        self._item_data = []

        self._left_highlighter = ConfigDiffHighlighter(
            self._ui.left_text_browser.document(), "yaml"
        )
        self._right_highlighter = ConfigDiffHighlighter(
            self._ui.right_text_browser.document(), "yaml"
        )
        self._ui.button_box.accepted.connect(self.close)

        self._ui.pkgs_list_widget.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self._ui.pkgs_list_widget.customContextMenuRequested.connect(
            self.on_pkg_context_menu_requested
        )

        # set up changed left element connection
        self._ui.pkgs_list_widget.currentItemChanged.connect(self._on_item_changed)

    def _init_pkg_context_menu(self):
        """Initalize context menu with all actions"""
        self.select_cntx_menu = QMenu()

        self.set_ref_item = QAction("Set as reference", self)
        self.set_themed_icon(self.set_ref_item, "icons/copy_link.svg")
        self.select_cntx_menu.addAction(self.set_ref_item)
        self.set_ref_item.triggered.connect(self._set_ref_item)

    def on_pkg_context_menu_requested(self, position):
        self.select_cntx_menu.exec(self._ui.pkgs_list_widget.mapToGlobal(position))

    @override
    def showEvent(self, event: QShowEvent) -> None:
        self._populate_left_items()
        self.update_diff()
        return super().showEvent(event)

    # public methods

    def add_diff_item(self, content: ConanPkg):
        """ " """
        self._item_data.append(content)

    def update_diff(self):
        """Resets the syntax highlighting, adds the different category items
        to the SyntaxHighlighter and then redraws them"""
        try:
            # reset diffs
            self._left_highlighter._reset_diff()
            self._right_highlighter._reset_diff()

            # set left and right content with first two diff elements
            pkg_diffs = list(diff(self._left_content, self._right_content))
            # filter other id
            pkg_diffs = list(filter(lambda diff: diff[1] != "id", pkg_diffs))
            for pkg_diff in pkg_diffs:
                diff_mode = pkg_diff[0]
                if diff_mode == "change":
                    if isinstance(pkg_diff[1], list):
                        key = str(pkg_diff[-1][1].split(".")[-1])
                    else:
                        key = str(pkg_diff[1].split(".")[-1])
                    value_left = str(pkg_diff[2][0])
                    value_right = str(pkg_diff[2][1])
                elif diff_mode == "add":
                    for detail_diff in pkg_diff[2]:
                        key = str(detail_diff[0])
                        value_right = str(detail_diff[1])
                        value_left = "INVALID"
                        self._left_highlighter.added_diffs.append(f"({key}: {value_left})")
                        self._right_highlighter.added_diffs.append(f"({key}: {value_right})")
                elif diff_mode == "remove":
                    for detail_diff in pkg_diff[2]:
                        key = str(detail_diff[0])
                        value_left = str(detail_diff[1])
                        value_right = "INVALID"
                        self._left_highlighter.removed_diffs.append(f"({key}: {value_left})")
                        self._right_highlighter.removed_diffs.append(f"({key}: {value_right})")
                self._left_highlighter.modified_diffs.append(f"({key}: {value_left})")
                self._right_highlighter.modified_diffs.append(f"({key}: {value_right})")
            self._left_highlighter.rehighlight()
            self._right_highlighter.rehighlight()
        except Exception as e:
            Logger().error("Diffing crashed: " + str(e))

    # internals

    def _on_item_changed(self, item: QListWidgetItem):
        """Set right content, when selection changes"""
        if not item:
            return
        sel_item_id = item.data(0)
        if "*" in sel_item_id:
            self._set_right_content(self._left_content)
            self.update_diff()
            return
        for item_data in self._item_data:
            if sel_item_id == item_data.get("id", ""):
                content = item_data
                self._set_right_content(content)
                self.update_diff()
                return

    def _filter_display_content(self, content: ConanPkg):
        """Only allow the following keys in the view"""
        new_content = {}
        new_content["settings"] = content.get("settings")
        new_content["options"] = content.get("options")
        new_content["requires"] = content.get("requires")
        return new_content

    def _set_left_content(self, content):
        self._left_content = self._filter_display_content(content)
        pkg_info = pformat(self._left_content).translate(
            {ord("{"): None, ord("}"): None, ord("'"): None}
        )
        self._ui.left_text_browser.setText(pkg_info)

    def _set_right_content(self, content):
        self._right_content = self._filter_display_content(content)
        pkg_info = pformat(self._right_content).translate(
            {ord("{"): None, ord("}"): None, ord("'"): None}
        )
        self._ui.right_text_browser.setText(pkg_info)

    def _set_ref_item(self):
        """Change the reference (*, left content)"""
        items = self._ui.pkgs_list_widget.selectedItems()
        if len(items) != 1:
            return
        sel_item = items[0]
        pkg_id = sel_item.data(0)
        for item in self._item_data:
            if item.get("id", "") == pkg_id:
                idx = self._item_data.index(item)
                self._item_data.pop(idx)
                self._item_data.insert(0, item)
                self._set_left_content(item)
                break
        self._ui.pkgs_list_widget.clear()
        self._populate_left_items()
        self.update_diff()

    def _populate_left_items(self):
        ref = self._set_diff_list_prios()

        self._set_left_content(ref)
        QListWidgetItem("* " + ref.get("id", "Unknown"), self._ui.pkgs_list_widget)
        for prio in range(0, 5):
            for content in self._item_data:
                if prio == content["prio"]:
                    item_name = content.get("id", "Unknown")
                    QListWidgetItem(item_name, self._ui.pkgs_list_widget)

        self._set_right_content(self._item_data[1])
        # select 2nd item so we see a diff per default
        self._ui.pkgs_list_widget.setCurrentRow(1)

    def _set_diff_list_prios(self):
        """Prioritize the list elements depending on the similarity of the original package
        Supports the most common settings.
        No option support yet.
        """
        ref = self._item_data[0]
        ref["prio"] = -1

        item_data_len = len(self._item_data)
        for i in range(1, item_data_len):
            content = self._item_data[i]
            pkg_diffs = list(diff(ref, content))
            pkg_arch_diffs = list(filter(lambda diff: diff[1] == "settings.arch", pkg_diffs))
            content["prio"] = 5
            if pkg_arch_diffs:
                content["prio"] = 4
                continue
            pkg_os_diffs = list(filter(lambda diff: diff[1] == "settings.os", pkg_diffs))
            if pkg_os_diffs:
                content["prio"] = 3
                continue
            pkg_compiler_diffs = list(
                filter(lambda diff: diff[1] == "settings.compiler", pkg_diffs)
            )
            if pkg_compiler_diffs:
                content["prio"] = 2
                continue
            pkg_compiler_version_diffs = list(
                filter(lambda diff: diff[1] == ["settings", "compiler.version"], pkg_diffs)
            )
            if pkg_compiler_version_diffs:
                content["prio"] = 1
                continue
            if not pkg_compiler_version_diffs:
                content["prio"] = 0
        return ref



import argparse
import os
import platform
import sys
from tempfile import gettempdir
from typing import TYPE_CHECKING

from conan_explorer import (
    APP_NAME,
    PKG_NAME,
    SETTINGS_FILE_NAME,
    __version__,
    asset_path,
    user_save_path,
)
from conan_explorer.settings import SETTINGS_INI_TYPE, settings_factory

from .base_ui.crash import bug_dialog_exc_hook
from .base_ui.loading import LoaderGui
from .base_ui.theming import activate_theme
from .logger import Logger
from .system import check_for_wayland

if TYPE_CHECKING:
    from conan_explorer.conan_wrapper import ConanCommonUnifiedApi, ConanWorker

### Global variables ###

active_settings = settings_factory(
    SETTINGS_INI_TYPE, user_save_path / (SETTINGS_FILE_NAME + ".ini")
)
conan_api: "ConanCommonUnifiedApi"  # initialized by load_conan
conan_worker: "ConanWorker"  # initialized by load_conan
qt_platform = ""


def run_application():
    """Start the Qt application and load the main window"""
    parse_cmd_args()
    init_platform()
    # Change cwd to temp, so temporary files become writable,
    # even if the current folder is not - Fixes #168
    os.chdir(gettempdir())

    qt_app = load_qapp()
    # Loading dialog until Conan is available
    loader = LoaderGui(None)
    loader.load_for_blocking(
        None,
        load_conan,
        (loader,),
        cancel_button=False,
        loading_text="Starting " + APP_NAME + " " + __version__,
    )
    loader.wait_for_finished()

    from PySide6 import QtGui

    from conan_explorer.ui.main_window import MainWindow

    # inline imports to optimize load times
    main_window = MainWindow(qt_app)
    main_window.setWindowIcon(QtGui.QIcon(str(asset_path / "icons" / "icon.ico")))
    main_window.show()  # show first, then load appsgrid with progress bar
    main_window.load()

    qt_app.exec()

    main_window.save_window_state()
    # cancel conan worker tasks on exit
    if conan_worker:
        conan_worker.finish_working(10)


def parse_cmd_args():
    """
    All CLI related functions.
    """
    parser = argparse.ArgumentParser(
        prog=PKG_NAME, description=f"{PKG_NAME} command line interface"
    )
    parser.add_argument("-v", "--version", action="version", version=__version__)
    parser.add_argument("-p", "--platform", type=str)

    args = parser.parse_args()
    global qt_platform
    qt_platform = args.platform


def init_platform():
    if platform.system() == "Windows":
        # Workaround for Windows, so that on the taskbar the
        # correct icon will be shown (and not the default python icon).
        from ctypes import windll  # Only exists on Windows.

        MY_APP_ID = "ConanAppLauncher." + __version__
        windll.shell32.SetCurrentProcessExplicitAppUserModelID(MY_APP_ID)
    elif platform.system() == "Darwin":
        print("Mac OS is currently not supported.")
        sys.exit(1)


def load_conan(loader: LoaderGui):
    global conan_api, conan_worker
    from conan_explorer.conan_wrapper import ConanApiFactory, ConanWorker

    conan_api = ConanApiFactory()
    conan_worker = ConanWorker(conan_api, active_settings)
    loader.loading_string_signal.emit("Initializing Conan")
    conan_api.init_api()


def load_qapp():
    """Load bootstrapping to be able to display a first widget."""
    # this import takes seconds - it could only be possibly parallelized
    # with more basics gui frameworks
    from PySide6 import QtCore, QtWidgets

    # Apply Qt attributes (only possible before QApplication is created)
    try:
        # Passthrough seems to work well for high dpi scaling
        QtWidgets.QApplication.setHighDpiScaleFactorRoundingPolicy(
            QtCore.Qt.HighDpiScaleFactorRoundingPolicy.PassThrough
        )
    except Exception:
        Logger().debug("Can't set DPI Rounding", exc_info=True)

    if qt_platform:
        os.environ["QT_QPA_PLATFORM"] = qt_platform
    else:
        if check_for_wayland():  # enable native Wayland support
            os.environ["QT_QPA_PLATFORM"] = "wayland"

    # to use icons in qss file
    QtCore.QDir.addSearchPath("icons", os.path.join(asset_path, "icons"))

    # disable reacting on light and dark themed mode automatically, otherwise
    # we can get get dark mode window bar in light mode
    QtWidgets.QApplication.setDesktopSettingsAware(False)

    qt_app = QtWidgets.QApplication(sys.argv)
    qt_app.setApplicationName(APP_NAME)
    qt_app.setApplicationDisplayName(APP_NAME)

    # Overwrite the excepthook with our own -
    # this will provide a method to report bugs for the user
    sys.excepthook = bug_dialog_exc_hook  # dialog needs qt_app
    activate_theme(qt_app)
    return qt_app



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'crash.ui'
##
## Created by: Qt User Interface Compiler version 6.6.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QApplication, QDialog, QDialogButtonBox,
    QLabel, QSizePolicy, QTextBrowser, QVBoxLayout,
    QWidget)

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(423, 291)
        self.verticalLayout = QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.crash_message_label = QLabel(Dialog)
        self.crash_message_label.setObjectName(u"crash_message_label")
        self.crash_message_label.setWordWrap(True)
        self.crash_message_label.setOpenExternalLinks(True)
        self.crash_message_label.setTextInteractionFlags(Qt.LinksAccessibleByMouse|Qt.TextSelectableByKeyboard)

        self.verticalLayout.addWidget(self.crash_message_label)

        self.error_text_browser = QTextBrowser(Dialog)
        self.error_text_browser.setObjectName(u"error_text_browser")
        self.error_text_browser.setReadOnly(True)
        self.error_text_browser.setOpenExternalLinks(True)

        self.verticalLayout.addWidget(self.error_text_browser)

        self.button_box = QDialogButtonBox(Dialog)
        self.button_box.setObjectName(u"button_box")
        self.button_box.setOrientation(Qt.Horizontal)
        self.button_box.setStandardButtons(QDialogButtonBox.Ok)

        self.verticalLayout.addWidget(self.button_box)


        self.retranslateUi(Dialog)
        self.button_box.accepted.connect(Dialog.accept)
        self.button_box.rejected.connect(Dialog.reject)

        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"Application Crash - Bug Report", None))
        self.crash_message_label.setText(QCoreApplication.translate("Dialog", u"Error text", None))
        self.error_text_browser.setHtml(QCoreApplication.translate("Dialog", u"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><meta charset=\"utf-8\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"hr { height: 1px; border-width: 0; }\n"
"li.unchecked::marker { content: \"\\2610\"; }\n"
"li.checked::marker { content: \"\\2612\"; }\n"
"</style></head><body style=\" font-family:'Segoe UI'; font-size:9pt; font-weight:400; font-style:normal;\">\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><br /></p></body></html>", None))
    # retranslateUi




from typing import Callable, List, Optional, Union

from PySide6.QtCore import (
    QAbstractItemModel,
    QModelIndex,
    QPersistentModelIndex,
    Qt,
    SignalInstance,
)
from PySide6.QtWidgets import QFileSystemModel
from typing_extensions import override

from conan_explorer.app.logger import Logger

QAF = Qt.AlignmentFlag
QORI = Qt.Orientation
QIDR = Qt.ItemDataRole
QINDEXLIKE = Union[QModelIndex, QPersistentModelIndex]


def re_register_signal(signal: SignalInstance, slot: Callable):
    try:  # need to be removed, otherwise will be called multiple times
        signal.disconnect()
    except RuntimeError:
        # no way to check if it is connected and it will throw an error
        pass
    signal.connect(slot)


class FileSystemModel(QFileSystemModel):
    """This fixes an issue with the header not being centered vertically"""

    def __init__(
        self, h_align=QAF.AlignLeft | QAF.AlignVCenter, v_align=QAF.AlignVCenter, parent=None
    ):
        super().__init__(parent)
        self.alignments = {QORI.Horizontal: h_align, QORI.Vertical: v_align}

    @override
    def headerData(self, section, orientation, role=0):
        if role == QIDR.TextAlignmentRole:
            return self.alignments[orientation]
        elif role == QIDR.DecorationRole:
            return None
        else:
            return QFileSystemModel.headerData(self, section, orientation, role)


class TreeModelItem:
    """
    Represents a tree item for a model view with lazy loading.
    Implemented like the default QT example.
    """

    def __init__(
        self, data: List[str], parent: Optional["TreeModelItem"] = None, lazy_loading=False
    ):
        self.parent_item = parent
        self.item_data = data
        self.child_items = []
        self.is_loaded = not lazy_loading

    def append_child(self, item):
        self.child_items.append(item)

    def remove_child(self, item):
        self.child_items.remove(item)

    def get_child_item_row(self, item):
        return self.child_items.index(item)

    def child(self, row):
        try:
            return self.child_items[row]
        except Exception:
            return None

    def child_count(self):
        return len(self.child_items)

    def column_count(self):
        return len(self.item_data)

    def data(self, column):
        return self.item_data[column]

    def parent(self):
        return self.parent_item

    def row(self):
        if self.parent_item:
            return self.parent_item.child_items.index(self)
        return 0

    def load_children(self):
        self.child_items = []
        self.is_loaded = True


class TreeModel(QAbstractItemModel):
    """Qt tree model to be used with TreeModelItem.
    Supports lazy loading, if TreeModelItem enables it."""

    def __init__(self, checkable=False, *args, **kwargs):
        super(TreeModel, self).__init__(*args, **kwargs)
        self.root_item = TreeModelItem([])
        self._checkable = checkable

    def clear_items(self):
        self.beginResetModel()
        self.root_item.child_items.clear()
        self.endResetModel()

    def add_item(self, item: TreeModelItem):  # to root_item
        child_count = self.root_item.child_count()
        item.parent_item = self.root_item
        self.beginInsertColumns(QModelIndex(), child_count, child_count)
        self.root_item.append_child(item)
        self.endInsertRows()

    def remove_item(self, item: TreeModelItem):  # from root_item
        item_index = self.get_index_from_item(item)
        self.beginRemoveRows(item_index.parent(), item_index.row(), item_index.row())
        self.root_item.remove_child(item)
        self.endRemoveRows()

    @override
    def columnCount(self, parent=None):
        if parent.isValid():
            return parent.internalPointer().column_count()  # type: ignore
        else:
            return self.root_item.column_count()

    @override
    def index(self, row, column, parent=None):
        if not self.hasIndex(row, column, parent):
            return QModelIndex()

        if not parent.isValid():
            parent_item = self.root_item
        else:
            parent_item = parent.internalPointer()

        child_item = parent_item.child(row)  # type: ignore
        if child_item:
            return self.createIndex(row, column, child_item)
        else:
            return QModelIndex()

    @override
    def data(self, index, role=0):
        raise NotImplementedError

    @override
    def rowCount(self, parent=None):
        if parent.column() > 0:
            return 0

        if not parent.isValid():
            parent_item = self.root_item
        else:
            parent_item = parent.internalPointer()

        return parent_item.child_count()  # type: ignore

    @override
    def flags(self, index):
        if not index.isValid():
            return Qt.ItemFlag.NoItemFlags
        flags = Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable
        if self._checkable:
            flags = flags | Qt.ItemFlag.ItemIsUserCheckable
        return flags

    @override
    def parent(self, child: QINDEXLIKE):  # type: ignore
        if not child.isValid():
            return QModelIndex()

        child_item = child.internalPointer()
        if not hasattr(child_item, "parent"):
            Logger().error("%s has incorrect type:", child_item)
        parent_item = child_item.parent()

        if not parent_item or parent_item == self.root_item:
            return QModelIndex()

        return self.createIndex(parent_item.row(), 0, parent_item)

    @override
    def headerData(self, section, orientation, role=0):
        if orientation == Qt.Orientation.Horizontal and role == QIDR.DisplayRole:
            return self.root_item.data(section)
        return None

    @override
    def canFetchMore(self, parent):
        if not parent.isValid():
            return False
        item = parent.internalPointer()
        return not item.is_loaded  # enabled, if lazy loading is enabled

    @override
    def fetchMore(self, parent):
        item: TreeModelItem = parent.internalPointer()
        item.load_children()

    def get_index_from_item(self, item: TreeModelItem) -> QModelIndex:
        # find the row with the matching reference
        found_item = False
        ref_row = 0
        for ref_row in range(self.root_item.child_count()):
            current_item = self.root_item.child_items[ref_row]
            # always has one dummy child count
            for child_row in range(len(current_item.child_items)):
                current_child_item = current_item.child_items[child_row]
                if current_child_item == item:
                    found_item = True
                    parent_index = self.index(ref_row, 0, QModelIndex())
                    return self.index(child_row, 0, parent_index)
            if current_item == item:
                found_item = True
                return self.index(ref_row, 0, QModelIndex())
        if not found_item:
            Logger().debug("Cannot find %s in model", str(item))
            return QModelIndex()
        return self.index(ref_row, 0, QModelIndex())



import json
from dataclasses import asdict
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar

import jsonschema
from packaging.version import Version

from conan_explorer.app.typing import SignatureCheckMeta

from . import (UiAppGridConfig, UiAppLinkConfig, UiConfig, UiConfigInterface,
               UiTabConfig)

if TYPE_CHECKING:
    from typing import TypedDict
else:
    try:
        from typing import TypedDict
    except ImportError:
        from typing_extensions import TypedDict

from conan_explorer import PathLike, asset_path
from conan_explorer.app.logger import Logger

# Internal represantation of JSON save format


class JsonAppConfig(TypedDict):
    """ Internal representation of json file format"""
    version: str
    tabs: List[Dict]  # same as ConfigTypes, but as dict


class ConanOptionConfig(TypedDict):
    """ Internal representation of conan options in json file format"""
    name: str
    value: str


class JsonUiConfig(UiConfigInterface, metaclass=SignatureCheckMeta):

    def __init__(self, json_file_path: PathLike):
        self._json_file_path = Path(json_file_path)

        # create file, if not available for first start
        if not self._json_file_path.exists():
            Logger().info("Quicklaunch: Creating json file")
            self._json_file_path.touch()
        else:
            Logger().info(f"Quicklaunch: Using {self._json_file_path}")

    @classmethod
    def get_file_ext(cls):
        return ".json"

    T = TypeVar('T', bound="UiTabConfig | UiAppLinkConfig")
    @staticmethod
    def _convert_to_config_type(dict: Dict[str, Any], config_type: Type[T]) -> T:
        """ Convert the dict to the class representations. 
        The matching is 1-to-1 except the conan_options and that Apps
        must instantiate the appropriate class
        """
        result_config = config_type()
        for key in dict.keys():  # matches 1-1
            value = dict[key]
            if key == "apps":
                value = []
                for dict_val in dict[key]:
                    app = JsonUiConfig._convert_to_config_type(dict_val, UiAppLinkConfig)
                    value.append(app)
            elif key == "conan_options":
                options: List[ConanOptionConfig] = dict[key]
                value = {}
                for option in options:
                    value[option["name"]] = option.get("value", "")
            setattr(result_config, key, value)
        return result_config

    def load(self) -> UiConfig:
        """ Parse the json config file, validate and convert to object structure """
        json_app_config: JsonAppConfig = {"version": "0.0.0", "tabs": []}
        Logger().debug("Quicklaunch: Loading file '%s'...", self._json_file_path)

        with open(asset_path / "config_schema.json") as schema_file:
            json_schema = json.load(schema_file)

        with open(str(self._json_file_path)) as fp:
            if fp.read():
                fp.seek(0)
                try:
                    json_app_config = json.load(fp)
                    jsonschema.validate(instance=json_app_config, schema=json_schema)
                except Exception as e:
                    Logger().error(f"Quicklaunch: Failed validating:\n{str(e)}")
                    return UiConfig()

        # implement subsequent migration functions
        self.migrate_to_0_3_0(json_app_config)
        self.migrate_to_0_4_0(json_app_config)

        # build the object abstraction and update
        tabs_result: List[UiTabConfig] = []
        for tab_dict in json_app_config.get("tabs", {}):
            tab_result = self._convert_to_config_type(tab_dict, UiTabConfig)
            tabs_result.append(tab_result)

        # auto update version to last version
        json_app_config["version"] = json_schema.get("properties").get("version").get("enum")[-1]

        # write it back with updates
        with open(str(self._json_file_path), "w") as config_file:
            json.dump(json_app_config, config_file, indent=4)
        return UiConfig(UiAppGridConfig(tabs=tabs_result))

    def save(self, app_config: UiConfig):
        """ Create json dict from model and write it to path. """
        tabs = app_config.app_grid.tabs
        tabs_data = []
        for tab in tabs:
            tab_dict = asdict(tab)
            # convert options and conan ref to a json writable format
            for app_dict in tab_dict.get("apps", {}):
                opt_list = []
                for opt_key in app_dict.get("conan_options", {}):
                    opt_list.append({"name": opt_key, "value": app_dict["conan_options"][opt_key]})
                app_dict["conan_options"] = opt_list
            tabs_data.append(tab_dict)

        # get last version
        with open(asset_path / "config_schema.json") as schema_file:
            json_schema = json.load(schema_file)
        version = json_schema.get("properties").get("version").get("enum")[-1]
        json_app_config: JsonAppConfig = {"version": version, "tabs": tabs_data}

        with open(str(self._json_file_path), "w") as config_file:
            json.dump(json_app_config, config_file, indent=2)

    def migrate_to_0_3_0(self, app_config: JsonAppConfig):
        """ Compatibility function to update schema version. """
        if Version(app_config["version"]) >= Version("0.3.0"):
            return
        for tab in app_config["tabs"]:
            for app in tab["apps"]:
                if app.get("package_id"):
                    value = app.pop("package_id")
                    app["conan_ref"] = value

    def migrate_to_0_4_0(self, app_config: JsonAppConfig):
        """ Compatibility function to update schema version. """
        if Version(app_config["version"]) >= Version("0.4.0"):
            return
        for tab in app_config["tabs"]:
            for app in tab["apps"]:
                if app.get("console_application"):
                    value = app.pop("console_application")
                    app["is_console_application"] = value



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'package_explorer.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QAbstractScrollArea, QApplication, QFrame,
    QHBoxLayout, QHeaderView, QLabel, QLayout,
    QLineEdit, QPushButton, QSizePolicy, QSplitter,
    QTabWidget, QTextBrowser, QTextEdit, QTreeView,
    QVBoxLayout, QWidget)

class Ui_Form(object):
    def setupUi(self, Form):
        if not Form.objectName():
            Form.setObjectName(u"Form")
        Form.resize(600, 379)
        self.verticalLayout = QVBoxLayout(Form)
        self.verticalLayout.setSpacing(4)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 3)
        self.local_packages_bar_layout = QHBoxLayout()
        self.local_packages_bar_layout.setObjectName(u"local_packages_bar_layout")
        self.local_packages_bar_layout.setSizeConstraint(QLayout.SizeConstraint.SetMinimumSize)
        self.local_packages_bar_layout.setContentsMargins(-1, -1, 7, -1)
        self.refresh_button = QPushButton(Form)
        self.refresh_button.setObjectName(u"refresh_button")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.refresh_button.sizePolicy().hasHeightForWidth())
        self.refresh_button.setSizePolicy(sizePolicy)
        self.refresh_button.setMinimumSize(QSize(32, 32))
        self.refresh_button.setMaximumSize(QSize(32, 32))

        self.local_packages_bar_layout.addWidget(self.refresh_button)

        self.show_sizes_button = QPushButton(Form)
        self.show_sizes_button.setObjectName(u"show_sizes_button")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.show_sizes_button.sizePolicy().hasHeightForWidth())
        self.show_sizes_button.setSizePolicy(sizePolicy1)
        self.show_sizes_button.setMinimumSize(QSize(32, 32))
        self.show_sizes_button.setMaximumSize(QSize(32, 32))
        self.show_sizes_button.setCheckable(True)

        self.local_packages_bar_layout.addWidget(self.show_sizes_button)

        self.expand_all_button = QPushButton(Form)
        self.expand_all_button.setObjectName(u"expand_all_button")
        self.expand_all_button.setMinimumSize(QSize(32, 32))
        self.expand_all_button.setMaximumSize(QSize(32, 32))

        self.local_packages_bar_layout.addWidget(self.expand_all_button)

        self.package_filter_label = QLabel(Form)
        self.package_filter_label.setObjectName(u"package_filter_label")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Preferred)
        sizePolicy2.setHorizontalStretch(1)
        sizePolicy2.setVerticalStretch(1)
        sizePolicy2.setHeightForWidth(self.package_filter_label.sizePolicy().hasHeightForWidth())
        self.package_filter_label.setSizePolicy(sizePolicy2)
        self.package_filter_label.setMinimumSize(QSize(0, 32))
        self.package_filter_label.setMaximumSize(QSize(16777215, 32))

        self.local_packages_bar_layout.addWidget(self.package_filter_label)

        self.splitter_filter = QSplitter(Form)
        self.splitter_filter.setObjectName(u"splitter_filter")
        sizePolicy3 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)
        sizePolicy3.setHorizontalStretch(1)
        sizePolicy3.setVerticalStretch(1)
        sizePolicy3.setHeightForWidth(self.splitter_filter.sizePolicy().hasHeightForWidth())
        self.splitter_filter.setSizePolicy(sizePolicy3)
        self.splitter_filter.setOrientation(Qt.Orientation.Horizontal)
        self.splitter_filter.setHandleWidth(5)
        self.splitter_filter.setChildrenCollapsible(False)
        self.package_filter_edit = QLineEdit(self.splitter_filter)
        self.package_filter_edit.setObjectName(u"package_filter_edit")
        self.package_filter_edit.setEnabled(True)
        sizePolicy4 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Minimum)
        sizePolicy4.setHorizontalStretch(1)
        sizePolicy4.setVerticalStretch(1)
        sizePolicy4.setHeightForWidth(self.package_filter_edit.sizePolicy().hasHeightForWidth())
        self.package_filter_edit.setSizePolicy(sizePolicy4)
        self.package_filter_edit.setMinimumSize(QSize(150, 32))
        self.package_filter_edit.setMaximumSize(QSize(350, 32))
        font = QFont()
        font.setPointSize(10)
        self.package_filter_edit.setFont(font)
        self.package_filter_edit.setInputMethodHints(Qt.InputMethodHint.ImhNone)
        self.package_filter_edit.setClearButtonEnabled(True)
        self.splitter_filter.addWidget(self.package_filter_edit)
        self.package_path_label = QTextBrowser(self.splitter_filter)
        self.package_path_label.setObjectName(u"package_path_label")
        sizePolicy5 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        sizePolicy5.setHorizontalStretch(0)
        sizePolicy5.setVerticalStretch(0)
        sizePolicy5.setHeightForWidth(self.package_path_label.sizePolicy().hasHeightForWidth())
        self.package_path_label.setSizePolicy(sizePolicy5)
        self.package_path_label.setMaximumSize(QSize(16777215, 32))
        self.package_path_label.setStyleSheet(u"text-align: right;")
        self.package_path_label.setFrameShape(QFrame.Shape.NoFrame)
        self.package_path_label.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.package_path_label.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.package_path_label.setSizeAdjustPolicy(QAbstractScrollArea.SizeAdjustPolicy.AdjustToContents)
        self.package_path_label.setAutoFormatting(QTextEdit.AutoFormattingFlag.AutoNone)
        self.package_path_label.setLineWrapMode(QTextEdit.LineWrapMode.NoWrap)
        self.package_path_label.setOverwriteMode(False)
        self.splitter_filter.addWidget(self.package_path_label)

        self.local_packages_bar_layout.addWidget(self.splitter_filter)


        self.verticalLayout.addLayout(self.local_packages_bar_layout)

        self.splitter = QSplitter(Form)
        self.splitter.setObjectName(u"splitter")
        sizePolicy1.setHeightForWidth(self.splitter.sizePolicy().hasHeightForWidth())
        self.splitter.setSizePolicy(sizePolicy1)
        self.splitter.setOrientation(Qt.Orientation.Horizontal)
        self.splitter.setChildrenCollapsible(False)
        self.package_select_view = QTreeView(self.splitter)
        self.package_select_view.setObjectName(u"package_select_view")
        sizePolicy6 = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Expanding)
        sizePolicy6.setHorizontalStretch(1)
        sizePolicy6.setVerticalStretch(1)
        sizePolicy6.setHeightForWidth(self.package_select_view.sizePolicy().hasHeightForWidth())
        self.package_select_view.setSizePolicy(sizePolicy6)
        self.package_select_view.setMinimumSize(QSize(100, 0))
        self.package_select_view.setFrameShape(QFrame.Shape.NoFrame)
        self.package_select_view.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.package_select_view.setSizeAdjustPolicy(QAbstractScrollArea.SizeAdjustPolicy.AdjustToContents)
        self.package_select_view.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.package_select_view.setIndentation(15)
        self.package_select_view.setUniformRowHeights(True)
        self.package_select_view.setSortingEnabled(True)
        self.package_select_view.setAnimated(True)
        self.package_select_view.setAllColumnsShowFocus(True)
        self.package_select_view.setWordWrap(True)
        self.package_select_view.setHeaderHidden(False)
        self.splitter.addWidget(self.package_select_view)
        self.package_select_view.header().setVisible(True)
        self.package_tab_widget = QTabWidget(self.splitter)
        self.package_tab_widget.setObjectName(u"package_tab_widget")
        sizePolicy7 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy7.setHorizontalStretch(2)
        sizePolicy7.setVerticalStretch(0)
        sizePolicy7.setHeightForWidth(self.package_tab_widget.sizePolicy().hasHeightForWidth())
        self.package_tab_widget.setSizePolicy(sizePolicy7)
        self.package_tab_widget.setMinimumSize(QSize(0, 0))
        self.package_tab_widget.setDocumentMode(False)
        self.package_tab_widget.setTabsClosable(True)
        self.package_tab_widget.setMovable(False)
        self.package_tab_widget.setTabBarAutoHide(False)
        self.tab = QWidget()
        self.tab.setObjectName(u"tab")
        sizePolicy8 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.MinimumExpanding)
        sizePolicy8.setHorizontalStretch(2)
        sizePolicy8.setVerticalStretch(0)
        sizePolicy8.setHeightForWidth(self.tab.sizePolicy().hasHeightForWidth())
        self.tab.setSizePolicy(sizePolicy8)
        self.horizontalLayout = QHBoxLayout(self.tab)
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.package_file_view = QTreeView(self.tab)
        self.package_file_view.setObjectName(u"package_file_view")
        sizePolicy9 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Expanding)
        sizePolicy9.setHorizontalStretch(3)
        sizePolicy9.setVerticalStretch(1)
        sizePolicy9.setHeightForWidth(self.package_file_view.sizePolicy().hasHeightForWidth())
        self.package_file_view.setSizePolicy(sizePolicy9)
        self.package_file_view.setMinimumSize(QSize(100, 0))
        self.package_file_view.setFrameShape(QFrame.Shape.NoFrame)
        self.package_file_view.setSizeAdjustPolicy(QAbstractScrollArea.SizeAdjustPolicy.AdjustToContents)
        self.package_file_view.setEditTriggers(QAbstractItemView.EditTrigger.EditKeyPressed|QAbstractItemView.EditTrigger.SelectedClicked)
        self.package_file_view.setTabKeyNavigation(False)
        self.package_file_view.setDragDropMode(QAbstractItemView.DragDropMode.DragDrop)
        self.package_file_view.setDefaultDropAction(Qt.DropAction.TargetMoveAction)
        self.package_file_view.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.package_file_view.setIndentation(15)
        self.package_file_view.setUniformRowHeights(True)
        self.package_file_view.setItemsExpandable(True)
        self.package_file_view.setSortingEnabled(True)
        self.package_file_view.setAnimated(True)
        self.package_file_view.setAllColumnsShowFocus(True)

        self.horizontalLayout.addWidget(self.package_file_view)

        self.package_tab_widget.addTab(self.tab, "")
        self.add_new_tab_tab = QWidget()
        self.add_new_tab_tab.setObjectName(u"add_new_tab_tab")
        self.package_tab_widget.addTab(self.add_new_tab_tab, "")
        self.splitter.addWidget(self.package_tab_widget)

        self.verticalLayout.addWidget(self.splitter)

        QWidget.setTabOrder(self.package_select_view, self.package_tab_widget)
        QWidget.setTabOrder(self.package_tab_widget, self.package_file_view)
        QWidget.setTabOrder(self.package_file_view, self.refresh_button)
        QWidget.setTabOrder(self.refresh_button, self.show_sizes_button)
        QWidget.setTabOrder(self.show_sizes_button, self.package_filter_edit)
        QWidget.setTabOrder(self.package_filter_edit, self.package_path_label)

        self.retranslateUi(Form)

        self.package_tab_widget.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(Form)
    # setupUi

    def retranslateUi(self, Form):
        Form.setWindowTitle(QCoreApplication.translate("Form", u"Form", None))
#if QT_CONFIG(tooltip)
        self.refresh_button.setToolTip(QCoreApplication.translate("Form", u"Refresh package list view", None))
#endif // QT_CONFIG(tooltip)
        self.refresh_button.setText("")
#if QT_CONFIG(tooltip)
        self.show_sizes_button.setToolTip(QCoreApplication.translate("Form", u"Show package sizes", None))
#endif // QT_CONFIG(tooltip)
        self.show_sizes_button.setText("")
#if QT_CONFIG(tooltip)
        self.expand_all_button.setToolTip(QCoreApplication.translate("Form", u"Expand all/Collapse all", None))
#endif // QT_CONFIG(tooltip)
        self.expand_all_button.setText("")
        self.package_filter_label.setText(QCoreApplication.translate("Form", u"Filter:", None))
        self.package_filter_edit.setPlaceholderText(QCoreApplication.translate("Form", u"*", None))
        self.package_path_label.setHtml(QCoreApplication.translate("Form", u"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><meta charset=\"utf-8\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"hr { height: 1px; border-width: 0; }\n"
"li.unchecked::marker { content: \"\\2610\"; }\n"
"li.checked::marker { content: \"\\2612\"; }\n"
"</style></head><body style=\" font-family:'Segoe UI'; font-size:9pt; font-weight:400; font-style:normal;\">\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><br /></p></body></html>", None))
        self.package_tab_widget.setTabText(self.package_tab_widget.indexOf(self.tab), QCoreApplication.translate("Form", u"New tab", None))
        self.package_tab_widget.setTabText(self.package_tab_widget.indexOf(self.add_new_tab_tab), QCoreApplication.translate("Form", u"  +  ", None))
    # retranslateUi





""" Classes representing the ui config (Data Transfer Objects) 
this is a bit overblown for this usecase, but it is worth an experiment """

from abc import abstractmethod
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Dict, List

from conan_explorer import INVALID_CONAN_REF, PathLike

UI_CONFIG_JSON_TYPE = "json"

if TYPE_CHECKING:
    from ..model import UiAppLinkModel, UiTabModel


@dataclass
class UiAppLinkConfig():
    name: str = "New App"
    executable: str = ""
    icon: str = ""
    is_console_application: bool = False
    args: str = ""
    conan_options: Dict[str, str] = field(default_factory=dict)
    conan_ref: str = INVALID_CONAN_REF


@dataclass
class UiTabConfig():
    name: str = "New Tab"
    # The Union is a workaround. How to say, that this is the base class?
    apps: List["UiAppLinkConfig | UiAppLinkModel"] = field(default_factory=list)

    def __post_init__(self):
        if not self.apps:
            self.apps.append(UiAppLinkConfig())


@dataclass
class UiAppGridConfig():
    # The Union is a workaround. How to say, that this is the base class?
    tabs: List["UiTabConfig | UiTabModel"] = field(default_factory=list)

    def __post_init__(self):
        if not self.tabs:
            self.tabs.append(UiTabConfig())


@dataclass
class UiConfig():
    app_grid: UiAppGridConfig = field(default_factory=UiAppGridConfig)


def ui_config_factory(type: str, source: PathLike) -> "UiConfigInterface":

    if type == UI_CONFIG_JSON_TYPE:
        from .json_file import JsonUiConfig
        implementation = JsonUiConfig(source)
    else:
        raise NotImplementedError
    return implementation


def get_ui_config_file_ext(type: str) -> str:
    if type == UI_CONFIG_JSON_TYPE:
        from .json_file import JsonUiConfig
        file_ext = JsonUiConfig.get_file_ext()
    else:
        raise NotImplementedError
    return file_ext


class UiConfigInterface():
    """
    Abstract Class to implement ui config mechanism.
    Source artefact should be passed by constructor an and is not changeable.
    """

    @classmethod
    def get_file_ext(cls) -> str:  # enter file extesnion like .yaml
        raise NotImplementedError

    @abstractmethod
    def load(self) -> UiConfig:
        raise NotImplementedError

    @abstractmethod
    def save(self, app_config: UiConfig):
        raise NotImplementedError



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'remote_login_dialog.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QAbstractItemView, QAbstractScrollArea, QApplication,
    QDialog, QDialogButtonBox, QGridLayout, QLabel,
    QLayout, QLineEdit, QListView, QListWidget,
    QListWidgetItem, QSizePolicy, QVBoxLayout, QWidget)

from conan_explorer.ui.widgets.password_line_edit import PasswordLineEdit

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(429, 273)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(Dialog.sizePolicy().hasHeightForWidth())
        Dialog.setSizePolicy(sizePolicy)
        Dialog.setModal(True)
        self.verticalLayout = QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(-1, -1, 0, 0)
        self.edit_grid = QGridLayout()
        self.edit_grid.setObjectName(u"edit_grid")
        self.edit_grid.setSizeConstraint(QLayout.SetMinAndMaxSize)
        self.edit_grid.setContentsMargins(-1, -1, 4, -1)
        self.password_label = QLabel(Dialog)
        self.password_label.setObjectName(u"password_label")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.password_label.sizePolicy().hasHeightForWidth())
        self.password_label.setSizePolicy(sizePolicy1)
        self.password_label.setMinimumSize(QSize(0, 0))
        self.password_label.setWordWrap(True)

        self.edit_grid.addWidget(self.password_label, 2, 0, 1, 1)

        self.name_label = QLabel(Dialog)
        self.name_label.setObjectName(u"name_label")
        self.name_label.setWordWrap(True)

        self.edit_grid.addWidget(self.name_label, 1, 0, 1, 1)

        self.name_line_edit = QLineEdit(Dialog)
        self.name_line_edit.setObjectName(u"name_line_edit")

        self.edit_grid.addWidget(self.name_line_edit, 1, 1, 1, 1)

        self.password_line_edit = PasswordLineEdit(Dialog)
        self.password_line_edit.setObjectName(u"password_line_edit")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.password_line_edit.sizePolicy().hasHeightForWidth())
        self.password_line_edit.setSizePolicy(sizePolicy2)
        self.password_line_edit.setMinimumSize(QSize(0, 0))
        self.password_line_edit.setFrame(True)
        self.password_line_edit.setEchoMode(QLineEdit.Password)
        self.password_line_edit.setClearButtonEnabled(False)

        self.edit_grid.addWidget(self.password_line_edit, 2, 1, 1, 1)

        self.remote_list = QListWidget(Dialog)
        self.remote_list.setObjectName(u"remote_list")
        sizePolicy3 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.MinimumExpanding)
        sizePolicy3.setHorizontalStretch(0)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.remote_list.sizePolicy().hasHeightForWidth())
        self.remote_list.setSizePolicy(sizePolicy3)
        self.remote_list.setMinimumSize(QSize(0, 40))
        self.remote_list.setSizeAdjustPolicy(QAbstractScrollArea.AdjustToContentsOnFirstShow)
        self.remote_list.setSelectionMode(QAbstractItemView.NoSelection)
        self.remote_list.setMovement(QListView.Static)
        self.remote_list.setFlow(QListView.TopToBottom)
        self.remote_list.setProperty("isWrapping", False)
        self.remote_list.setResizeMode(QListView.Adjust)
        self.remote_list.setUniformItemSizes(True)
        self.remote_list.setWordWrap(True)
        self.remote_list.setSelectionRectVisible(True)
        self.remote_list.setSortingEnabled(False)

        self.edit_grid.addWidget(self.remote_list, 0, 0, 1, 2)


        self.verticalLayout.addLayout(self.edit_grid)

        self.note_label = QLabel(Dialog)
        self.note_label.setObjectName(u"note_label")

        self.verticalLayout.addWidget(self.note_label)

        self.button_box = QDialogButtonBox(Dialog)
        self.button_box.setObjectName(u"button_box")
        sizePolicy4 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        sizePolicy4.setHorizontalStretch(1)
        sizePolicy4.setVerticalStretch(1)
        sizePolicy4.setHeightForWidth(self.button_box.sizePolicy().hasHeightForWidth())
        self.button_box.setSizePolicy(sizePolicy4)
        self.button_box.setOrientation(Qt.Horizontal)
        self.button_box.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.verticalLayout.addWidget(self.button_box)

        self.verticalLayout.setStretch(0, 1)

        self.retranslateUi(Dialog)
        self.button_box.rejected.connect(Dialog.reject)

        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"Login to Remote", None))
        self.password_label.setText(QCoreApplication.translate("Dialog", u"Password", None))
        self.name_label.setText(QCoreApplication.translate("Dialog", u"Name", None))
        self.note_label.setText("")
    # retranslateUi




from pathlib import Path
from typing import Dict, Optional, Protocol, Tuple, TypedDict, Union, runtime_checkable

from PySide6.QtCore import QSize
from PySide6.QtGui import QIcon, QImage, QPixmap
from PySide6.QtWidgets import QWidget

import conan_explorer.app as app

# reimport for usage in other files
from conan_explorer.app.base_ui.theming import get_gui_dark_mode
from conan_explorer.app.logger import Logger
from conan_explorer.settings import GUI_STYLE, GUI_STYLE_FLUENT, GUI_STYLE_MATERIAL
from conan_explorer.ui.common import (
    draw_svg_with_color,
    get_icon_from_image_file,
    get_inverted_asset_image,
)


@runtime_checkable
class CanSetIconWidgetProtocol(Protocol):
    def setIcon(self, icon: Union[QIcon, QPixmap]) -> None: ...


@runtime_checkable
class CanSetPixmapWidgetProtocol(Protocol):
    def setPixmap(self, arg__1: Union[QPixmap, QImage]) -> None: ...


def get_gui_style():
    gui_style = app.active_settings.get_string(GUI_STYLE).lower()
    if gui_style not in [GUI_STYLE_FLUENT, GUI_STYLE_MATERIAL]:
        Logger().warning("Theming: Can't read GUI style, setting it to material.")
        app.active_settings.set(GUI_STYLE, GUI_STYLE_MATERIAL)
        return GUI_STYLE_MATERIAL
    return gui_style


def get_asset_image_path(image_path: str) -> Path:
    asset_path = Path(image_path)
    if asset_path.exists():  # absolute path - return immediately
        return asset_path

    asset_path = app.asset_path / image_path
    if not asset_path.exists():  # try in style
        asset_path = asset_path.parent / get_gui_style() / asset_path.name

    if not asset_path.exists():
        Logger().warning(f"Can't find image: {str(asset_path)}")
        # set to an existing placeholder image to avoid crashes
        asset_path = app.asset_path / "icons/global/conan_logo.svg"
    return asset_path


def get_themed_asset_icon(
    image_path: str, force_light_mode=False, force_dark_mode=False
) -> QIcon:
    asset_path = get_asset_image_path(image_path)
    if force_dark_mode or (get_gui_dark_mode() and not force_light_mode):
        if asset_path.suffix == ".svg":
            asset_path = draw_svg_with_color(asset_path)
        else:
            asset_path = get_inverted_asset_image(asset_path)
    return get_icon_from_image_file(asset_path)


class ThemedWidget(QWidget):
    class IconInfo(TypedDict):
        asset_path: str
        size: Optional[Tuple[int, int]]

    def __init__(self, parent=None) -> None:
        if parent is not None:  # signals, it is already a widget
            super().__init__(parent)
        self._icon_map: Dict[
            Union[CanSetIconWidgetProtocol, CanSetPixmapWidgetProtocol], ThemedWidget.IconInfo
        ] = {}  # widget: {name, size} for re-theming

    def set_themed_icon(
        self,
        widget: Union[CanSetIconWidgetProtocol, CanSetPixmapWidgetProtocol],
        asset_path: str,
        size: Optional[Tuple[int, int]] = None,
        force_dark_mode=False,
        force_light_mode=False,
    ):
        """
        Applies an icon to a widget and inverts it, when theming is toggled to dark mode.
        For that reload_themed_icons must be called.
        Size only applies for pixmaps.
        """
        icon = get_themed_asset_icon(asset_path, force_light_mode, force_dark_mode)
        if isinstance(widget, CanSetIconWidgetProtocol):
            widget.setIcon(icon)
        elif isinstance(widget, CanSetPixmapWidgetProtocol):
            if size is None:
                size = (20, 20)
            widget.setPixmap(icon.pixmap(*size))
        self._icon_map[widget] = {"asset_path": asset_path, "size": size}

    def reload_themed_icons(self):
        for widget, info in self._icon_map.items():
            asset_rel_path = info["asset_path"]
            size = info["size"]
            icon = get_themed_asset_icon(asset_rel_path)
            if isinstance(widget, CanSetIconWidgetProtocol):
                widget.setIcon(icon)
            elif isinstance(widget, CanSetPixmapWidgetProtocol):
                widget.setPixmap(icon.pixmap(QSize(*size)))



"""Qt Loading dialog. Placed outside of ui, because
it is needed to boostrap the loading of all uis."""

import os
import time
from datetime import datetime
from time import sleep
from typing import Any, Callable, Optional, Tuple

from PySide6 import QtGui
from PySide6.QtCore import QObject, Qt, QThread, Signal, SignalInstance
from PySide6.QtWidgets import QApplication, QProgressDialog, QPushButton, QWidget

from conan_explorer import DEBUG_LEVEL, asset_path
from conan_explorer.app.logger import Logger
from conan_explorer.app.system import str2bool


class Worker(QObject):
    """Generic worker for Qt, which can call any function with args"""

    finished: SignalInstance = Signal(object)  # type: ignore

    def __init__(self, func, args: Tuple[Any, ...] = ()):
        super().__init__()
        self.func = func
        self.args = args

    def work(self):
        if DEBUG_LEVEL > 1:  # pragma: no cover
            try:
                import debugpy  # - debug with this the Qt Thread

                debugpy.debug_this_thread()
            except Exception:
                Logger().debug("Debugger not loaded!")
        start = time.time()
        ret = self.func(*self.args)
        end = time.time()
        Logger().debug("Waited: %.4f s", end - start)
        self.finished.emit(ret)


class LoaderGui(QObject):
    # Reuse the same progress_dialog instance for every loading dialog -
    # there can only be one at a time
    __progress_dialog: Optional[QProgressDialog] = None
    loading_string_signal: SignalInstance = Signal(str)  # type: ignore
    loading_gui_signal: SignalInstance = Signal(SignalInstance)  # type: ignore
    loading_finished_signal: SignalInstance = Signal()  # type: ignore
    default_loading_text = "Loading..."

    def __init__(self, parent: Optional[QObject]):
        super().__init__(parent)
        # initial setup
        if LoaderGui.__progress_dialog is None:  # implicit check for init
            wt = Qt.WindowType
            progress_dialog = QProgressDialog()
            progress_dialog.setAttribute(Qt.WidgetAttribute.WA_ShowModal)
            progress_dialog.setAttribute(Qt.WidgetAttribute.WA_AlwaysStackOnTop)
            # Window flags to disable close button
            progress_dialog.setWindowFlags(
                wt.FramelessWindowHint | wt.Window | wt.WindowTitleHint | wt.CustomizeWindowHint
            )
            progress_dialog.setCancelButton(None)  # type: ignore
            progress_dialog.setModal(True)  # user could trigger it twice -> crash
            progress_dialog.setRange(0, 0)
            progress_dialog.setMinimumDuration(1000)
            progress_dialog.setMinimumWidth(500)
            progress_dialog.setMaximumWidth(600)
            progress_dialog.setWindowIcon(QtGui.QIcon(str(asset_path / "icons" / "icon.ico")))

            LoaderGui.__progress_dialog = progress_dialog
        self.progress_dialog = LoaderGui.__progress_dialog
        self.loading_string_signal.connect(self.progress_dialog.setLabelText)
        self.loading_finished_signal.connect(self.thread_finished)
        self.worker: Optional[Worker] = None
        self.load_thread: Optional[QThread] = None
        self.finished = True
        self.return_value = None

    def load_with_finish_hook(
        self,
        dialog_parent: Optional[QWidget],
        work_task: Callable,
        worker_args: Tuple[Any, ...] = (),
        finish_task: Optional[Callable] = None,
        loading_text: str = "",
        cancel_button=True,
    ):
        self.load(
            dialog_parent, work_task, worker_args, finish_task, loading_text, cancel_button
        )

    def load_for_blocking(
        self,
        dialog_parent: Optional[QWidget],
        work_task: Callable,
        worker_args: Tuple[Any, ...] = (),
        loading_text: str = "",
        cancel_button=True,
    ):
        self.load(
            dialog_parent,
            work_task,
            worker_args,
            loading_text=loading_text,
            cancel_button=cancel_button,
        )

    def load(
        self,
        dialog_parent: Optional[QWidget],
        work_task: Callable,
        worker_args: Tuple[Any, ...] = (),
        finish_task: Optional[Callable] = None,
        loading_text: str = "",
        cancel_button=True,
    ):
        self.finished = False
        self.return_value = None
        if not loading_text:
            loading_text = self.default_loading_text
        self.progress_dialog.setLabelText(loading_text)

        if cancel_button:
            self.cancel_button = QPushButton("Cancel")
            self.progress_dialog.setCancelButton(self.cancel_button)
            self.cancel_button.clicked.connect(self.on_cancel)
        else:
            self.progress_dialog.setCancelButton(None)  # type: ignore

        # set position in middle of window
        qapp: QApplication = QApplication.instance()  # type: ignore
        rectangle = self.progress_dialog.frameGeometry()
        if dialog_parent:
            # wait for active window to center
            start_time = datetime.now()
            while not qapp.activeWindow():
                QApplication.processEvents()
                time_delta = datetime.now() - start_time
                if time_delta.total_seconds() >= 2:
                    break
            if qapp.activeWindow():
                rectangle.moveCenter(qapp.activeWindow().frameGeometry().center())
                self.progress_dialog.move(rectangle.topLeft())
        else:
            rectangle.moveCenter(qapp.primaryScreen().geometry().center())
            self.progress_dialog.move(rectangle.topLeft())
        self.progress_dialog.show()
        self.progress_dialog.activateWindow()

        # for debug purposes only
        if str2bool(os.getenv("DISABLE_ASYNC_LOADER", "")):
            ret = work_task(*worker_args)
            self.thread_finished()
            self.progress_dialog.hide()
            if finish_task:
                try:  # difficult to handle without Qt signals - so just try
                    finish_task(ret)
                except Exception:
                    finish_task()
            self.on_finished(ret)
            return

        self.worker = Worker(work_task, worker_args)
        self.load_thread = QThread(dialog_parent)
        self.load_thread.setObjectName(f"loader_thread_{str(self)}")
        self.worker.moveToThread(self.load_thread)
        self.load_thread.started.connect(self.worker.work)

        if finish_task:
            self.worker.finished.connect(finish_task)
        self.load_thread.finished.connect(self.thread_finished)

        self.worker.finished.connect(self.worker.deleteLater)
        self.worker.finished.connect(self.load_thread.quit)
        self.worker.finished.connect(self.on_finished)

        self.load_thread.finished.connect(self.load_thread.deleteLater)
        self.load_thread.finished.connect(self.progress_dialog.hide)

        Logger().debug("Start async loading thread for %s", str(work_task))
        self.load_thread.start()
        if cancel_button:
            self.cancel_button.clicked.connect(self.load_thread.quit)

    def thread_finished(self):
        self.finished = True

    def wait_for_finished(self):
        Logger().debug("Wait for loading thread...")
        # execute once
        while not self.finished:
            sleep(0.01)
            QApplication.processEvents()

    def on_cancel(self):
        self.finished = True

    def on_finished(self, return_value=None):
        # save return value to get after wait_for_finished
        self.return_value = return_value




from pathlib import Path
from typing import Optional

import conan_explorer.app as app  # using global module pattern
from conan_explorer import (DEFAULT_UI_CFG_FILE_NAME,
                                LEGACY_UI_CFG_FILE_NAME, legacy_user_save_path,
                                user_save_path)
from conan_explorer.settings import AUTO_INSTALL_QUICKLAUNCH_REFS, LAST_CONFIG_FILE
from . import (UI_CONFIG_JSON_TYPE, UiConfig,
                        UiConfigInterface, get_ui_config_file_ext, ui_config_factory)

from ..model import UiAppGridModel


class UiApplicationModel(UiConfig):
    CONFIG_TYPE = UI_CONFIG_JSON_TYPE

    def __init__(self, conan_pkg_installed=None, *args, **kwargs):
        """ Create an empty AppModel on init, so we can load it later"""
        UiConfig.__init__(self, *args, **kwargs)
        self.app_grid: UiAppGridModel
        self._ui_config_data: Optional[UiConfigInterface] = None
        self.conan_pkg_installed = conan_pkg_installed

    def save(self):
        """ Save configuration """
        if self._ui_config_data:
            self._ui_config_data.save(self)

    def load(self, ui_config: UiConfig) -> "UiApplicationModel":
        """ Load the model and submodels from the config object """
        # update conan info
        self.app_grid = UiAppGridModel()
        self.app_grid.load(ui_config.app_grid, self)
        if app.conan_worker and app.active_settings.get_bool(AUTO_INSTALL_QUICKLAUNCH_REFS):
            app.conan_worker.finish_working(3)
            app.conan_worker.update_all_info(self.app_grid.get_all_conan_worker_elements(),
                                             self.emit_conan_pkg_signal_callback)

        return self

    def emit_conan_pkg_signal_callback(self, conan_ref: str, pkg_id: str):
        if not self.conan_pkg_installed:
            return
        self.conan_pkg_installed.emit(conan_ref, pkg_id)

    def loadf(self, config_source: str) -> "UiApplicationModel":
        """ Load model and submodels from specified file source """
        # empty ui config, create it in user path
        file_ext = get_ui_config_file_ext(self.CONFIG_TYPE)
        default_config_file_path = user_save_path / (DEFAULT_UI_CFG_FILE_NAME + file_ext)
        if not config_source or legacy_user_save_path / (LEGACY_UI_CFG_FILE_NAME) == Path(config_source):
            config_source = str(default_config_file_path)
            app.active_settings.set(LAST_CONFIG_FILE, str(config_source))

        self._ui_config_data = ui_config_factory(self.CONFIG_TYPE, config_source)
        ui_config = self._ui_config_data.load()

        self.load(ui_config)
        return self




import datetime
from dataclasses import dataclass
from time import sleep
from typing import Optional

from PySide6.QtCore import QRect, Qt, Signal, SignalInstance
from PySide6.QtGui import QDesktopServices, QKeySequence, QResizeEvent, QShortcut
from PySide6.QtWidgets import (
    QApplication,
    QFileDialog,
    QFrame,
    QPushButton,
    QRadioButton,
    QVBoxLayout,
)
from typing_extensions import override

import conan_explorer.app as app  # using global module pattern
from conan_explorer import (
    APP_NAME,
    DEBUG_LEVEL,
    ENABLE_GUI_STYLES,
    MAX_FONT_SIZE,
    MIN_FONT_SIZE,
    PathLike,
    conan_version,
)
from conan_explorer.app import LoaderGui, Logger, activate_theme
from conan_explorer.settings import (
    AUTO_OPEN_LAST_VIEW,
    CONSOLE_SPLIT_SIZES,
    FILE_EDITOR_EXECUTABLE,
    FONT_SIZE,
    GUI_MODE,
    GUI_MODE_DARK,
    GUI_MODE_LIGHT,
    GUI_STYLE,
    GUI_STYLE_FLUENT,
    GUI_STYLE_MATERIAL,
    LAST_CONFIG_FILE,
    LAST_VIEW,
    WINDOW_SIZE,
)

from .common import init_qt_logger, remove_qt_logger
from .common.theming import get_gui_dark_mode, get_gui_style, get_themed_asset_icon
from .dialogs import FileEditorSelDialog
from .fluent_window import FluentWindow, SideSubMenu
from .plugin import PluginHandler, PluginInterfaceV1
from .views import AboutPage, AppGridView, PluginsPage
from .views.app_grid.config.model import UiApplicationModel
from .widgets import AnimatedToggle, WideMessageBox


# Signal names
@dataclass
class BaseSignals:
    """Dict of base signals, wich is passed down to the Views"""

    conan_pkg_installed: SignalInstance  # conan_ref, pkg_id
    conan_pkg_removed: SignalInstance  # conan_ref, pkg_ids
    conan_remotes_updated: SignalInstance


class MainWindow(FluentWindow):
    """Instantiates MainWindow and holds all UI objects"""

    # signals for inter page communication
    # conan_ref, pkg_ids
    conan_pkg_installed: SignalInstance = Signal(str, str)  # type: ignore
    # conan_ref, pkg_ids
    conan_pkg_removed: SignalInstance = Signal(str, str)  # type: ignore
    conan_remotes_updated: SignalInstance = Signal()  # type: ignore
    log_console_message: SignalInstance = Signal(str)  # type: ignore

    qt_logger_name = "qt_logger"
    _can_close = True  # wait for blocking operations

    def __init__(self, qt_app: QApplication):
        super().__init__(title_text=APP_NAME)
        self._qt_app = qt_app
        self.loaded = False
        self.base_signals = BaseSignals(
            self.conan_pkg_installed, self.conan_pkg_removed, self.conan_remotes_updated
        )
        self.model = UiApplicationModel(self.conan_pkg_installed)
        self._plugin_handler = PluginHandler(self, self.base_signals, self.page_widgets)
        self.setWindowTitle("")  # app display name is already there
        # connect logger to console widget to log possible errors at init
        init_qt_logger(Logger(), self.qt_logger_name, self.log_console_message)
        self.log_console_message.connect(self.write_log)
        # Default pages
        self.about_page = AboutPage(self, self.base_signals)
        self.plugins_page = PluginsPage(self, self._plugin_handler)
        self.app_grid = AppGridView(
            self, self.model.app_grid, self.base_signals, self.page_widgets
        )
        self._init_left_menu()
        self._init_right_menu()
        self.ui.title_icon_label.setPixmap(
            get_themed_asset_icon("icons/icon.ico", force_light_mode=True).pixmap(20, 20)
        )
        self._conan_minor_version = str(conan_version.major) + "." + str(conan_version.minor)
        self.ui.search_bar_line_edit.setPlaceholderText(
            f"Search Conan {self._conan_minor_version} docs"
        )

        for key in (
            "Enter",
            "Return",
        ):
            self._search_shorcut = QShortcut(
                key, self.ui.search_bar_line_edit, self.on_docs_searched
            )
            self._search_shorcut.setContext(Qt.ShortcutContext.WidgetWithChildrenShortcut)
        self._plugin_handler.load_plugin.connect(self._post_load_plugin)
        self._plugin_handler.unload_plugin.connect(self._unload_plugin)

        # size needs to be set as early as possible to correctly position loading windows
        self.restore_window_state()

    @override
    def close(self):
        while not self._can_close:
            sleep(0.1)
        return super().close()

    def on_docs_searched(self):
        extra_addr = ""
        if conan_version.major == 1:
            extra_addr = "en/"
        # https://docs.conan.io/en/1.62/search.html?q=abc&check_keywords=yes&area=default
        search_url = (
            f"https://docs.conan.io/{extra_addr}{self._conan_minor_version}/search.html"
            f"?q={self.ui.search_bar_line_edit.text()}&check_keywords=yes&area=default"
        )
        QDesktopServices.openUrl(search_url)

    def _init_left_menu(self):
        self.add_left_menu_entry(
            "Conan Quicklaunch",
            "icons/global/grid.svg",
            is_upper_menu=True,
            page_widget=self.app_grid,
            create_page_menu=True,
        )
        # set default page
        self.page_widgets.get_button_by_name("Conan Quicklaunch").click()
        if DEBUG_LEVEL > 0:
            self.add_left_menu_entry(
                "Reload", "icons/refresh.svg", False, page_widget=self.app_grid
            )
            button: QPushButton = self.findChild(QPushButton, "reload")  # type: ignore
            button.clicked.connect(self.reload)

    def _init_style_chooser(self):
        self._style_chooser_frame = QFrame(self)
        self._style_chooser_layout = QVBoxLayout(self._style_chooser_frame)
        self._style_chooser_radio_material = QRadioButton("Material", self._style_chooser_frame)
        self._style_chooser_layout.addWidget(self._style_chooser_radio_material)
        self._style_chooser_radio_fluent = QRadioButton("Fluent", self._style_chooser_frame)
        self._style_chooser_layout.addWidget(self._style_chooser_radio_fluent)
        # set initial state
        if get_gui_style() == GUI_STYLE_MATERIAL:
            self._style_chooser_radio_material.setChecked(True)
        elif get_gui_style() == GUI_STYLE_FLUENT:
            self._style_chooser_radio_fluent.setChecked(True)

        self._style_chooser_radio_material.clicked.connect(self.on_style_changed)
        self._style_chooser_radio_fluent.clicked.connect(self.on_style_changed)

    def _init_right_menu(self):
        self.main_general_settings_menu.add_button_menu_entry(
            "Select file editor",
            self.open_file_editor_selection_dialog,
            "icons/edit_file.svg",
        )
        view_settings_submenu = SideSubMenu(self.ui.right_menu_bottom_content_sw, "View")

        self.main_general_settings_menu.add_sub_menu(view_settings_submenu, "icons/view.svg")

        view_settings_submenu.add_button_menu_entry(
            "Font Size +",
            self.on_font_size_increased,
            "icons/increase_font.svg",
            QKeySequence("CTRL++"),
            self,
        )
        view_settings_submenu.add_button_menu_entry(
            "Font Size - ",
            self.on_font_size_decreased,
            "icons/decrease_font.svg",
            QKeySequence("CTRL+-"),
            self,
        )
        view_settings_submenu.add_menu_line()

        view_settings_submenu.add_toggle_menu_entry(
            "Dark Mode",
            self.on_dark_mode_changed,
            get_gui_dark_mode(),
            "icons/dark_mode.svg",
        )

        view_settings_submenu.add_toggle_menu_entry(
            "Open latest view",
            self.on_auto_open_last_view_changed,
            app.active_settings.get_bool(AUTO_OPEN_LAST_VIEW),
            "icons/refresh.svg",
        )
        if ENABLE_GUI_STYLES:
            self._init_style_chooser()
            view_settings_submenu.add_named_custom_entry(
                "Icon Style",
                self._style_chooser_frame,
                "icons/global/conan_settings.svg",
                force_v_layout=True,
            )
            view_settings_submenu.add_menu_line()

        self.main_general_settings_menu.add_menu_line()

        if conan_version.major == 1:
            self.main_general_settings_menu.add_button_menu_entry(
                "Remove Locks", self.on_conan_remove_locks, "icons/remove-lock.svg"
            )
            self.main_general_settings_menu.add_button_menu_entry(
                "Clean Conan Cache", self.open_cleanup_cache_dialog, "icons/cleanup.svg"
            )
            self.main_general_settings_menu.add_button_menu_entry(
                "Repair Package Metadata",
                self.open_repair_pkg_metadata_dialog,
                "icons/home_repair.svg",
            )
            self.main_general_settings_menu.add_menu_line()
        self.add_right_bottom_menu_main_page_entry(
            "Manage Plugins", self.plugins_page, "icons/plugin.svg"
        )
        self.add_right_bottom_menu_main_page_entry("About", self.about_page, "icons/about.svg")

    @override
    def closeEvent(self, event):
        """Remove qt logger, so it doesn't log into a non existant object"""
        self.app_grid.setEnabled(False)  # disable app_grid to signal shutdown
        try:
            self.log_console_message.disconnect(self.write_log)
        except Exception:
            # Sometimes the closeEvent is called twice and disconnect errors.
            pass
        remove_qt_logger(Logger(), self.qt_logger_name)
        super().closeEvent(event)

    # Try to fix random exit from maximized state
    # @override
    # def resizeEvent(self, event: QResizeEvent) -> None:
    #     super().resizeEvent(event)
    #     try:
    #         if self.loaded:
    #             self.ui.page_stacked_widget.currentWidget().setMaximumWidth(
    #                 self.ui.center_frame.width() - 4
    #             )
    #             self.ui.page_stacked_widget.currentWidget().adjustSize()
    #     except Exception as e:
    #         Logger().error(f"Can't resize current view: {str(e)}")

    def load(self, config_source: Optional[PathLike] = None):
        """Load all application gui elements specified in the GUI config (file)"""
        config_source_str = str(config_source)
        if not config_source:
            config_source_str = app.active_settings.get_string(LAST_CONFIG_FILE)
        self._load_plugins()  # creates the objects - must be in this thread

        loader = LoaderGui(self)
        loader.load_for_blocking(
            self, self._load_job, (config_source_str,), loading_text="Loading Plugins"
        )
        loader.wait_for_finished()
        # Restore last view
        if app.active_settings.get_bool(AUTO_OPEN_LAST_VIEW):
            try:
                last_view = app.active_settings.get_string(LAST_VIEW)
                page = self.page_widgets.get_page_by_name(last_view)
                self.page_widgets.get_button_by_type(type(page)).click()
            except Exception as e:
                Logger().debug("Can't switch to page for auto open: " + str(e))
                # click first view
                page = self.page_widgets.get_all_pages()[0]
                self.page_widgets.get_button_by_type(type(page)).click()

        self.loaded = True

    def _load_plugins(self):
        self._plugin_handler.load_all_plugins()

    def _load_job(self, config_source: str):
        self._plugin_handler.post_load_plugins()
        self._load_quicklaunch(config_source)

    def _post_load_plugin(self, plugin_object: PluginInterfaceV1):
        try:
            self.add_left_menu_entry(
                plugin_object.plugin_description.name,
                plugin_object.plugin_description.icon,
                True,
                plugin_object,
                plugin_object.plugin_description.side_menu,
            )
        except Exception as e:
            Logger().error(
                f"Can't load plugin {plugin_object.plugin_description.name}: {str(e)}"
            )

    def reload(self):
        """Only for debug mode"""
        self._plugin_handler.reload_all_plugins()
        # self.reload_theme()

    def _unload_plugin(self, plugin_name: str):
        self.page_widgets.remove_page_extras_by_name(plugin_name)

    def _load_quicklaunch(self, config_source: str):
        # load ui file definitions
        self.model.loadf(config_source)
        # now actually load the views - this need signals, to execute in the gui thread
        self.app_grid.model = self.model.app_grid
        self.app_grid.load_signal.emit()

    def on_conan_remove_locks(self):
        app.conan_api.remove_locks()

    def on_font_size_increased(self):
        """Increase font size by 2. Ignore if font gets too large."""
        new_size = app.active_settings.get_int(FONT_SIZE) + 1
        if new_size > MAX_FONT_SIZE:
            return
        app.active_settings.set(FONT_SIZE, new_size)
        activate_theme(self._qt_app)

    def on_font_size_decreased(self):
        """Decrease font size by 2. Ignore if font gets too small."""
        new_size = app.active_settings.get_int(FONT_SIZE) - 1
        if new_size < MIN_FONT_SIZE:
            return
        app.active_settings.set(FONT_SIZE, new_size)
        activate_theme(self._qt_app)

    def on_style_changed(self):
        # wait 0,5 seconds, so all animations can finish
        start = datetime.datetime.now()
        while datetime.datetime.now() - start <= datetime.timedelta(milliseconds=600):
            QApplication.processEvents()
        if self._style_chooser_radio_material.isChecked():
            app.active_settings.set(GUI_STYLE, GUI_STYLE_MATERIAL)
        elif self._style_chooser_radio_fluent.isChecked():
            app.active_settings.set(GUI_STYLE, GUI_STYLE_FLUENT)
        self.reload_theme()

    def on_dark_mode_changed(self):
        sender_toggle: AnimatedToggle = self.sender()  # type: ignore
        sender_toggle.wait_for_anim_finish()
        enable_dark_mode = sender_toggle.isChecked()
        if enable_dark_mode:
            app.active_settings.set(GUI_MODE, GUI_MODE_DARK)
        else:
            app.active_settings.set(GUI_MODE, GUI_MODE_LIGHT)
        self.reload_theme()

    def reload_theme(self):
        # This must run in the main thread!
        self._can_close = False
        activate_theme(self._qt_app)

        # all icons must be reloaded
        self.apply_theme()
        for page in self.page_widgets.get_all_pages():
            page.reload_themed_icons()
        self._can_close = True

    def on_auto_open_last_view_changed(self):
        sender_toggle: AnimatedToggle = self.sender()  # type: ignore
        sender_toggle.wait_for_anim_finish()
        app.active_settings.set(AUTO_OPEN_LAST_VIEW, sender_toggle.isChecked())

    def open_repair_pkg_metadata_dialog(self):
        """Open the message box to confirm deletion of invalid cache folders"""
        from conan_explorer.conan_wrapper.conan_cleanup import ConanCleanup

        cleaner = ConanCleanup()

        loader = LoaderGui(self)
        loader.load(
            self,
            cleaner.gather_invalid_remote_metadata,
        )
        loader.wait_for_finished()
        invalid_refs = cleaner.invalid_metadata_refs
        if not invalid_refs:
            Logger().info("No references with invalid metadata found!")

        msg_box = WideMessageBox(parent=self)
        button = WideMessageBox.StandardButton
        msg_box.setWindowTitle("Delete folders")
        msg_box.setText(
            (
                "Found the following packages with invalid package metadata. "
                "Attempt to repair? This can take a while, if you have many remotes. "
                "You can speed this up by disabling the remotes where you don't expect to "
                "find the missing references"
            )
        )
        msg_box.setDetailedText(str(invalid_refs))
        msg_box.setStandardButtons(button.Yes | button.Cancel)  # type: ignore
        msg_box.setIcon(WideMessageBox.Icon.Question)
        msg_box.setWidth(800)
        msg_box.setMaximumHeight(600)
        reply = msg_box.exec()
        if reply == button.Yes:

            def repair_refs(refs):
                for ref in refs:
                    cleaner.repair_invalid_remote_metadata(ref)
                    loader.loading_string_signal.emit("Trying to repair\n" + ref)

            loader.load(
                self, repair_refs, (invalid_refs,), loading_text="Repairing metadata..."
            )

    def open_cleanup_cache_dialog(self):
        """Open the message box to confirm deletion of invalid cache folders"""
        from .dialogs import ConanCacheCleanupDialog

        return ConanCacheCleanupDialog(self)

    def open_file_editor_selection_dialog(self):
        dialog = FileEditorSelDialog(self)
        if dialog.exec() == QFileDialog.DialogCode.Accepted:
            app.active_settings.set(FILE_EDITOR_EXECUTABLE, "")

    def write_log(self, text):
        """Write the text signaled by the logger"""
        self.ui.console.append(text)

    def restore_window_state(self):
        # restore window size
        sizes_str = app.active_settings.get_string(WINDOW_SIZE)
        if sizes_str.lower() == "maximized":
            self.showMaximized()
        else:
            split_sizes_int = [0, 0, 800, 600]
            try:
                split_sizes = sizes_str.strip().split(",")
                split_sizes_int = [int(size) for size in split_sizes]
            except Exception as e:
                Logger().warning(f"Can't read window size: {str(e)}")
            try:
                assert len(split_sizes_int) == 4, "Invalid setting window size length."
                geometry = QRect(*split_sizes_int)
                self.setGeometry(geometry)
            except Exception as e:
                Logger().warning(f"Can't restore window size: {str(e)}")

        # restore console size
        try:
            sizes_str = app.active_settings.get_string(CONSOLE_SPLIT_SIZES)
            split_sizes = sizes_str.strip().split(",")
            split_sizes_int = [int(size) for size in split_sizes]
            self.ui.content_footer_splitter.setSizes(split_sizes_int)
        except Exception as e:
            Logger().warning(f"Can't restore console size: {str(e)}")

    def save_window_state(self):
        # save window size
        if self.isMaximized():
            app.active_settings.set(WINDOW_SIZE, "maximized")
        else:
            geometry = self.geometry()
            geo_str = (
                f"{geometry.left()},{geometry.top()}," f"{geometry.width()},{geometry.height()}"
            )
            app.active_settings.set(WINDOW_SIZE, geo_str)
        # save console size
        sizes = self.ui.content_footer_splitter.sizes()
        if len(sizes) < 2:
            Logger().warning("Can't save splitter size")
        sizes_str = f"{int(sizes[0])},{int(sizes[1])}"
        app.active_settings.set(CONSOLE_SPLIT_SIZES, sizes_str)

        # save last view
        page: PluginInterfaceV1 = self.ui.page_stacked_widget.currentWidget()  # type: ignore
        try:
            app.active_settings.set(LAST_VIEW, page.plugin_description.name)
        except Exception:
            pass  # this can only happen if we close before any paged opened - no need to save
